# ruff: noqa: SLF001
import asyncio
import json
import re
import time
from collections import namedtuple
from datetime import datetime, timedelta, timezone
from itertools import chain, cycle
from unittest import mock

import boto3
import botocore
import pytest
from botocore.stub import ANY, Stubber
from dagster_aws.utils import ensure_dagster_aws_tests_import
from dagster_cloud.workspace.ecs.client import Client, EcsServiceError, Service

ensure_dagster_aws_tests_import()
from dagster_aws_tests.ecs_tests.stubbed_ecs import StubbedEcs


@pytest.fixture
def client(cluster_name="test"):
    client = Client(
        cluster_name=cluster_name,
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )

    def _mock_assign_public_ip(*args, **kwags):
        return "ENABLED"

    client._infer_assign_public_ip = _mock_assign_public_ip

    yield client


@pytest.fixture
def stubbed_ecs():
    return StubbedEcs(boto3.client("ecs"))


@pytest.fixture
def stubbed_client(stubbed_ecs, cluster_name="test"):
    client = Client(
        cluster_name=cluster_name,
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        ecs_client=stubbed_ecs,
    )

    def _mock_assign_public_ip(*args, **kwags):
        return "ENABLED"

    client._infer_assign_public_ip = _mock_assign_public_ip

    yield client


@pytest.fixture
def stubber(client):
    with Stubber(client.ecs) as stubber:
        yield stubber


@pytest.mark.parametrize("cluster_name", ["test", "arn:aws:ecs:region:012345678910:cluster/test"])
def test_cluster_name(client, cluster_name):
    assert client.cluster_name == "test"


def test_taggable(client, stubber):
    stubber.add_response(
        method="list_account_settings",
        service_response={
            "settings": [
                {"name": "serviceLongArnFormat", "value": "enabled"},
            ]
        },
    )
    assert client.taggable

    stubber.add_response(
        method="list_account_settings",
        service_response={
            "settings": [
                {"name": "serviceLongArnFormat", "value": "disabled"},
            ]
        },
    )
    assert not client.taggable


def test_create_service_tags(client, stubber):
    arn = "arn:aws:ecs:us-east-1:1234567890:service/cluster-name/service-name"
    tags = {"foo": "bar"}

    params_without_tags = [
        "clientToken",
        "cluster",
        "desiredCount",
        "launchType",
        "networkConfiguration",
        "serviceName",
        "serviceRegistries",
        "taskDefinition",
        "enableExecuteCommand",
        "propagateTags",
    ]

    # When the new ARN format is disabled
    stubber.add_response(
        method="list_account_settings",
        service_response={
            "settings": [
                {"name": "serviceLongArnFormat", "value": "disabled"},
            ]
        },
    )
    # Expect that tags don't get added
    expected_params_without_tags = dict.fromkeys(params_without_tags, ANY)
    stubber.add_response(
        method="create_service",
        service_response={"service": {"serviceArn": arn}},
        expected_params=expected_params_without_tags,
    )

    client._create_service(
        service_name="fake",
        service_registry_arn="fake",
        task_definition_arn="fake",
        tags=tags,
    )

    # When the new ARN format is enabled
    stubber.add_response(
        method="list_account_settings",
        service_response={
            "settings": [
                {"name": "serviceLongArnFormat", "value": "enabled"},
            ]
        },
    )
    # Expect that tags get added
    expected_params_with_tags = dict(
        expected_params_without_tags, tags=[{"key": "foo", "value": "bar"}]
    )
    stubber.add_response(
        method="create_service",
        service_response={"service": {"serviceArn": arn}},
        expected_params=expected_params_with_tags,
    )

    client._create_service(
        service_name="fake",
        service_registry_arn="fake",
        task_definition_arn="fake",
        tags=tags,
    )


def test_ecs_service_error():
    with pytest.raises(EcsServiceError) as ex:
        raise EcsServiceError(
            cluster="My cluster",
            task_arn="foo",
            stopped_reason="bar",
            logs=["My", "bad"],
            show_debug_prompt=False,
        )

    ex.match("ECS service failed because task foo failed: bar\n\nTask logs:\nMy\nbad")

    with pytest.raises(EcsServiceError) as ex:
        raise EcsServiceError(
            cluster="My cluster",
            task_arn="foo",
            stopped_reason="bar",
            logs=["My", "bad"],
            show_debug_prompt=True,
        )

    ex.match(
        "ECS service failed because task foo failed: bar\n\nTask logs:\nMy\nbad\n\nFor more"
        " information about the failure, check the ECS console for logs for task foo in cluster My"
        " cluster."
    )


def test_secrets(stubbed_client, stubbed_ecs):
    assert len(stubbed_ecs.list_task_definitions()["taskDefinitionArns"]) == 0

    secret_dict = {
        "secret_name1": "arn:aws:secretsmanager:us-west-2:111122223333:secret:aes128-1a2b3c",
        "secret_name2": "arn:aws:secretsmanager:us-west-2:111122223333:secret:aes192-4D5e6F",
    }

    stubbed_client.create_service(
        name="fake_service",
        image="fake_image",
        command=["fake_command"],
        execution_role_arn="fake-role",
        register_service_discovery=False,
        secrets=secret_dict,
    )

    task_arns = stubbed_ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_arns) == 1
    task_def = stubbed_ecs.describe_task_definition(taskDefinition=task_arns[0])

    assert task_def["taskDefinition"]["containerDefinitions"][0]["secrets"] == [
        {"name": secret_name, "valueFrom": secret_arn}
        for secret_name, secret_arn in secret_dict.items()
    ]


def test_wait_for_new_service_failure(client, stubber):
    service = Service(arn="arn:aws:ecs:region:012345678910:cluster/boom", client=client)

    failures = [
        {
            "arn": service.arn,
            "detail": "detail",
            "reason": "reason",
        },
    ]
    stubber.add_response(
        method="describe_services",
        service_response={
            "services": [],
            "failures": failures,
        },
    )

    with pytest.raises(
        Exception,
        match=re.escape(f"ECS DescribeServices API returned failures: {json.dumps(failures)}"),
    ):
        asyncio.run(client.wait_for_new_service(service, container_name=service.name))

    failures = [
        {
            "arn": service.arn,
            "detail": "detail1",
            "reason": "reason1",
        },
        {
            "arn": service.arn,
            "detail": "detail2",
            "reason": "reason2",
        },
    ]

    stubber.add_response(
        method="describe_services",
        service_response={
            "services": [],
            "failures": failures,
        },
    )
    with pytest.raises(
        Exception,
        match=re.escape(f"ECS DescribeServices API returned failures: {json.dumps(failures)}"),
    ) as ex:
        asyncio.run(client.wait_for_new_service(service, container_name=service.name))
    assert ex.match("detail1")
    assert ex.match("detail2")

    stubber.add_response(
        method="describe_services",
        service_response={
            "services": [],
            "failures": [],
        },
    )

    with pytest.raises(
        Exception,
        match=re.escape(
            f"ECS DescribeServices API returned an empty response for service {service.arn}."
        ),
    ):
        asyncio.run(client.wait_for_new_service(service, container_name=service.name))

    # Success after failure works due to grace period

    client.grace_period = 15

    async def _check_service_has_running_tasks(
        service_name, container_name, logger=None
    ) -> list[str]:
        return ["my_task"]

    client.check_service_has_running_tasks = _check_service_has_running_tasks

    stubber.add_response(
        method="describe_services",
        service_response={
            "services": [],
            "failures": failures,
        },
    )

    stubber.add_response(
        method="describe_services",
        service_response={
            "services": [
                {
                    "clusterArn": "arn:aws:ecs:us-east-1:012345678910:cluster/default",
                    "createdAt": time.time(),
                    "deploymentConfiguration": {
                        "maximumPercent": 200,
                        "minimumHealthyPercent": 100,
                    },
                    "deployments": [
                        {
                            "createdAt": time.time(),
                            "desiredCount": 1,
                            "id": "ecs-svc/9223370564341623665",
                            "pendingCount": 0,
                            "runningCount": 0,
                            "status": "PRIMARY",
                            "taskDefinition": (
                                "arn:aws:ecs:us-east-1:012345678910:task-definition/hello_world:6"
                            ),
                            "updatedAt": time.time(),
                        },
                    ],
                    "desiredCount": 1,
                    "events": [
                        {
                            "createdAt": time.time(),
                            "id": "38c285e5-d335-4b68-8b15-e46dedc8e88d",
                            # In this example, there is a service event that shows unavailable cluster resources.
                            "message": (
                                "(service ecs-simple-service) was unable to place a task because no"
                                " container instance met all of its requirements. The closest"
                                " matching (container-instance"
                                " 3f4de1c5-ffdd-4954-af7e-75b4be0c8841) is already using a port"
                                " required by your task. For more information, see the"
                                " Troubleshooting section of the Amazon ECS Developer Guide."
                            ),
                        },
                    ],
                    "loadBalancers": [],
                    "pendingCount": 0,
                    "runningCount": 0,
                    "serviceArn": "arn:aws:ecs:us-east-1:012345678910:service/ecs-simple-service",
                    "serviceName": "ecs-simple-service",
                    "status": "ACTIVE",
                    "taskDefinition": (
                        "arn:aws:ecs:us-east-1:012345678910:task-definition/hello_world:6"
                    ),
                },
            ],
            "failures": [],
        },
    )

    assert (
        asyncio.run(client.wait_for_new_service(service, container_name=service.name)) == "my_task"
    )


def test_check_service_has_running_task__success():
    with mock.patch("dagster_cloud.workspace.ecs.client.time") as mock_time:
        with mock.patch("dagster_cloud.workspace.ecs.client.asyncio.sleep") as mock_asyncio_sleep:
            mock_time.time.side_effect = chain([1, 20], cycle([30]))

            async def no_sleep():
                await asyncio.sleep(0)

            mock_asyncio_sleep.side_effect = no_sleep

            client = Client(
                cluster_name="test",
                service_discovery_namespace_id="fake-namespace",
                log_group="fake-log-group",
                grace_period=0,
            )

            # timeouts used by the method
            client.timeout = 60

            with Stubber(client.ecs) as stubber:
                stubber.add_response(
                    method="describe_services",
                    service_response={"services": [{"desiredCount": 2, "runningCount": 2}]},
                )
                stubber.add_response(
                    method="list_tasks", service_response={"taskArns": ["my_task", "your_task"]}
                )
                stubber.add_response(
                    method="describe_tasks",
                    service_response={
                        "tasks": [
                            {
                                "taskDefinitionArn": "arn:aws:ecs:us-east-1:012345678910:task-definition/my_service:x.y.z"
                            }
                        ]
                    },
                )
                client._check_for_stopped_tasks = mock.MagicMock(return_value=[])
                client._check_all_essential_containers_are_running = mock.MagicMock(
                    return_value=True
                )
                response = asyncio.run(
                    client.check_service_has_running_tasks("my_service", "my_container")
                )
                assert response == ["my_task", "your_task"]


@mock.patch("dagster_cloud.workspace.ecs.client.time")
def test_check_service_has_running_task__missing_running_tasks_timeout(mock_time):
    mock_time.time.side_effect = chain([1, 20], cycle([100]))
    mock_time.sleep.side_effect = lambda _: None

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )

    # timeouts used by the method
    client.timeout = 60

    with Stubber(client.ecs) as stubber:
        stubber.add_response(
            method="describe_services",
            service_response={"services": [{"desiredCount": 2, "runningCount": 1}]},
        )
        # Getting events from the service
        stubber.add_response(
            method="describe_services",
            service_response={"services": [{"events": [{"message": "kaput"}]}]},
        )

        with mock.patch.object(client, "_check_for_stopped_tasks", return_value=[]):
            with mock.patch.object(
                client, "_check_all_essential_containers_are_running", return_value=True
            ):
                with pytest.raises(
                    Exception,
                    match=r"Timed out waiting for a running task for service: my_service. Service events:\nkaput",
                ):
                    asyncio.run(
                        client.check_service_has_running_tasks("my_service", "my_container")
                    )


@mock.patch("dagster_cloud.workspace.ecs.client.time")
def test_check_service_has_running_task__stopped_tasks_after_grace_period(mock_time):
    mock_time.time.side_effect = chain([1, 32], cycle([1000]))
    mock_time.sleep.side_effect = lambda _: None

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )

    # timeouts used by the method
    client.timeout = 60

    with Stubber(client.ecs) as stubber:
        stubber.add_response(
            method="describe_services",
            service_response={"services": [{"desiredCount": 1, "runningCount": 0}]},
        )

        # todo: add a test for get_task_logs and mock it for this test
        stubber.add_response(
            method="describe_tasks",
            service_response={
                "tasks": [
                    {
                        "taskDefinitionArn": "arn:aws:ecs:us-east-1:012345678910:task-definition/my_service:x.y.z"
                    }
                ]
            },
        )
        stubber.add_response(
            method="describe_task_definition",
            service_response={
                "taskDefinition": {
                    "containerDefinitions": [
                        {
                            "name": "my_container",
                            "logConfiguration": {
                                "logDriver": "awslogs",
                                "options": {"awslogs-stream-prefix": "my_log_prefix"},
                            },
                        },
                    ]
                }
            },
        )
        with Stubber(client.logs) as logs_stubber:
            logs_stubber.add_response(
                method="get_log_events",
                service_response={"events": [{"message": "couch"}, {"message": "potato"}]},
            )

            with mock.patch.object(
                client,
                "_check_for_stopped_tasks",
                return_value=[
                    {
                        "name": "failed_task",
                        "createdAt": datetime.fromtimestamp(6),
                        "taskArn": "some_arn",
                        "stoppedReason": "kaput",
                    }
                ],
            ):
                with pytest.raises(
                    Exception,
                    match=r"ECS service failed because task some_arn failed: kaput\n\nTask logs:\ncouch\npotato",
                ):
                    asyncio.run(
                        client.check_service_has_running_tasks("my_service", "my_container")
                    )


@mock.patch("dagster_cloud.workspace.ecs.client.time")
def test_check_service_has_running_task__no_running_task_after_timeout(mock_time):
    mock_time.time.side_effect = chain([1, 30, 111], cycle([1000]))
    mock_time.sleep.side_effect = lambda _: None

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )

    # timeouts used by the method
    client.timeout = 60

    with Stubber(client.ecs) as stubber:
        # Waiting for tasks to be running
        stubber.add_response(
            method="describe_services",
            service_response={"services": [{"desiredCount": 1, "runningCount": 0}]},
        )
        # Getting events from the service
        stubber.add_response(
            method="describe_services",
            service_response={"services": [{"events": [{"message": "kaput"}]}]},
        )

        with mock.patch.object(client, "_check_for_stopped_tasks", return_value=[]):
            with pytest.raises(
                Exception,
                match=r"Timed out waiting for a running task for service: my_service. Service events:\nkaput",
            ):
                asyncio.run(client.check_service_has_running_tasks("my_service", "my_container"))


@mock.patch("dagster_cloud.workspace.ecs.client.time")
def test_check_service_has_running_task__task_stopped_with_transient_failure_immediately(mock_time):
    with mock.patch("dagster_cloud.workspace.ecs.client.time") as mock_time:
        with mock.patch("dagster_cloud.workspace.ecs.client.asyncio.sleep") as mock_asyncio_sleep:
            mock_time.time.side_effect = chain([1, 20], cycle([50]))

            async def no_sleep(_):
                return

            mock_asyncio_sleep.side_effect = no_sleep

            client = Client(
                cluster_name="test",
                service_discovery_namespace_id="fake-namespace",
                log_group="fake-log-group",
                grace_period=0,
            )

            # timeouts used by the method
            client.timeout = 60

            # simulate grace period for stopped tasks passing
            client.grace_period = 15

            with Stubber(client.ecs) as stubber:
                stubber.add_response(
                    method="describe_services",
                    service_response={"services": [{"desiredCount": 1, "runningCount": 0}]},
                )
                stubber.add_response(
                    method="list_tasks", service_response={"taskArns": ["my_task"]}
                )
                stubber.add_response(
                    method="describe_tasks",
                    service_response={
                        "tasks": [
                            {
                                "createdAt": datetime.fromtimestamp(6),
                                "taskArn": "some_arn",
                                "stoppedReason": "Timeout waiting for network interface provisioning to complete",
                                "lastStatus": "STOPPED",
                            }
                        ]
                    },
                )
                stubber.add_response(
                    method="describe_services",
                    service_response={"services": [{"desiredCount": 1, "runningCount": 1}]},
                )
                stubber.add_response(
                    method="list_tasks", service_response={"taskArns": ["your_task"]}
                )
                stubber.add_response(
                    method="describe_tasks",
                    service_response={
                        "tasks": [
                            {
                                "createdAt": datetime.fromtimestamp(6),
                                "taskArn": "new_arn",
                                "lastStatus": "RUNNING",
                            }
                        ]
                    },
                )
                client._check_all_essential_containers_are_running = mock.MagicMock(
                    return_value=True
                )

                response = asyncio.run(
                    client.check_service_has_running_tasks("my_service", "my_container")
                )
                assert response == ["your_task"]


@mock.patch("dagster_cloud.workspace.ecs.client.time")
def test_check_service_has_running_task__running_task_moves_to_transient_startup_failure(mock_time):
    with mock.patch("dagster_cloud.workspace.ecs.client.time") as mock_time:
        with mock.patch("dagster_cloud.workspace.ecs.client.asyncio.sleep") as mock_asyncio_sleep:
            mock_time.time.side_effect = chain([1, 20], cycle([30]))

            async def no_sleep(_):
                return

            mock_asyncio_sleep.side_effect = no_sleep

            client = Client(
                cluster_name="test",
                service_discovery_namespace_id="fake-namespace",
                log_group="fake-log-group",
                grace_period=0,
            )

            # timeouts used by the method
            client.timeout = 60

            with Stubber(client.ecs) as stubber:
                stubber.add_response(
                    method="describe_services",
                    service_response={"services": [{"desiredCount": 1, "runningCount": 1}]},
                )
                stubber.add_response(
                    method="list_tasks", service_response={"taskArns": ["my_task"]}
                )
                stubber.add_response(
                    method="describe_tasks",
                    service_response={
                        "tasks": [
                            {
                                "createdAt": datetime.fromtimestamp(6),
                                "taskArn": "some_arn",
                                "stoppedReason": "Timeout waiting for network interface provisioning to complete",
                                "lastStatus": "STOPPED",
                            }
                        ]
                    },
                )
                stubber.add_response(
                    method="describe_services",
                    service_response={"services": [{"desiredCount": 1, "runningCount": 1}]},
                )
                stubber.add_response(
                    method="list_tasks", service_response={"taskArns": ["your_task"]}
                )
                stubber.add_response(
                    method="describe_tasks",
                    service_response={
                        "tasks": [
                            {
                                "createdAt": datetime.fromtimestamp(6),
                                "taskArn": "new_arn",
                                "lastStatus": "RUNNING",
                            }
                        ]
                    },
                )
                client._check_for_stopped_tasks = mock.MagicMock(return_value=[])
                client._check_all_essential_containers_are_running = mock.MagicMock(
                    return_value=True
                )

                response = asyncio.run(
                    client.check_service_has_running_tasks("my_service", "my_container")
                )
                assert response == ["your_task"]


@pytest.mark.parametrize(
    "sub_test_name, containers, expected",
    [
        (
            "single_essential_container_running",
            [
                {
                    "name": "my_container",
                    "essential": True,
                    "lastStatus": "RUNNING",
                },
            ],
            True,
        ),
        (
            "single_essential_container_running_with_stopped_sidecar",
            [
                {
                    "name": "my_container",
                    "essential": True,
                    "lastStatus": "RUNNING",
                },
                {
                    "name": "my_sidecar",
                    "essential": False,
                    "lastStatus": "STOPPED",
                },
            ],
            True,
        ),
        (
            "single_essential_container_stopped",
            [
                {
                    "name": "my_container",
                    "essential": True,
                    "lastStatus": "STOPPED",
                },
            ],
            False,
        ),
        (
            "multiple_essential_containers_running",
            [
                {
                    "name": "my_container",
                    "essential": True,
                    "lastStatus": "RUNNING",
                },
                {
                    "name": "my_other_container",
                    "essential": True,
                    "lastStatus": "RUNNING",
                },
            ],
            True,
        ),
        (
            "multiple_essential_containers_with_one_stopped",
            [
                {
                    "name": "my_container",
                    "essential": True,
                    "lastStatus": "RUNNING",
                },
                {
                    "name": "my_other_container",
                    "essential": True,
                    "lastStatus": "STOPPED",
                },
            ],
            False,
        ),
    ],
)
def test_check_all_essential_containers_are_running(sub_test_name, containers, expected):
    # arrange
    mockTaskArn = "arn:aws:ecs:us-east-1:012345678910:task-definition/hello_world:x.y.z"
    task_definition = {
        "containerDefinitions": [
            {"name": c["name"], "essential": c["essential"]} for c in containers
        ]
    }

    task = {
        "taskDefinitionArn": mockTaskArn,
        "containers": [{"name": c["name"], "lastStatus": c["lastStatus"]} for c in containers],
    }

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )
    with mock.patch.object(
        client.ecs,
        "describe_task_definition",
        return_value={
            "taskDefinition": task_definition,
        },
    ) as mock_describe_task_definition:
        # act
        ret_val = client._check_all_essential_containers_are_running(task)

        # assert
        assert mock_describe_task_definition.call_count == 1
        mock_describe_task_definition.assert_called_with(taskDefinition=mockTaskArn)

        assert ret_val == expected


@pytest.mark.parametrize(
    "sub_test_name, stopped_tasks_arn, stopped_tasks, expected",
    [
        ("no_stopped_tasks", {"taskArns": []}, {"tasks": []}, []),
        (
            "three_stopped_task_sorted",
            {"taskArns": ["first", "third", "second"]},
            {
                "tasks": [
                    {
                        "name": "first",
                        "createdAt": datetime.fromtimestamp(1_000_000),
                    },
                    {
                        "name": "third",
                        "createdAt": datetime.fromtimestamp(3_000_000),
                    },
                    {
                        "name": "second",
                        "createdAt": datetime.fromtimestamp(2_000_000),
                    },
                ]
            },
            ["third", "second", "first"],  # sorted by createdAt in reverse order
        ),
    ],
)
def test_check_for_stopped_tasks(sub_test_name, stopped_tasks_arn, stopped_tasks, expected, caplog):
    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )
    with mock.patch.object(
        client.ecs, "list_tasks", return_value=stopped_tasks_arn
    ) as mock_list_tasks:
        ret_val = None
        if len(expected) > 1:
            with mock.patch.object(
                client.ecs, "describe_tasks", return_value=stopped_tasks
            ) as mock_describe_tasks:
                ret_val = client._check_for_stopped_tasks("my_service")
                assert mock_describe_tasks.call_count == 1
        else:
            ret_val = client._check_for_stopped_tasks("my_service")

        assert mock_list_tasks.call_count == 1
        assert [v["name"] for v in ret_val] == expected


def test_list_services(aws_mock, monkeypatch, caplog):
    monkeypatch.setattr(
        "dagster_cloud.workspace.ecs.client.Client.taggable",
        True,
    )
    monkeypatch.setattr(
        "dagster_cloud.workspace.ecs.client.Client._infer_assign_public_ip",
        lambda *args, **kwargs: "ENABLED",
    )
    caplog.set_level("WARNING")

    # Create VPC, subnet, and security group for moto 5 compatibility
    ec2 = boto3.client("ec2")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    vpc_id = vpc["Vpc"]["VpcId"]  # type: ignore
    subnet = ec2.create_subnet(VpcId=vpc_id, CidrBlock="10.0.1.0/24")
    subnet_id = subnet["Subnet"]["SubnetId"]  # type: ignore
    security_group = ec2.create_security_group(
        GroupName="test-sg", Description="Test security group", VpcId=vpc_id
    )
    security_group_id = security_group["GroupId"]

    ecs = boto3.client("ecs")

    ecs.create_cluster(clusterName="test")

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
        subnet_ids=[subnet_id],
        security_group_ids=[security_group_id],
    )

    client.create_service(
        name="tagged",
        image="test",
        command=["test"],
        execution_role_arn="test",
        tags={"foo": "bar"},
    )

    client.create_service(
        name="untagged",
        image="test",
        command=["test"],
        execution_role_arn="test",
    )

    assert len(client.list_services()) == 2

    # Backcompat: If we can't use the Resource Groups Tagging API
    # warn once and flip to legacy mode
    with mock.patch.object(
        client.tags_client,
        "get_paginator",
        side_effect=botocore.exceptions.ClientError(  # pyright: ignore[reportAttributeAccessIssue]
            operation_name="GetResources",
            error_response={"Error": {"Code": "AccessDeniedException"}},
        ),
    ):
        assert len(client.list_services(tags={"foo": "bar"})) == 1
        assert client.list_services(tags={"foo": "bar"})[0].name == "tagged"
        assert len(caplog.records) == 1

    # Re-initialize the client and enable the Resource Groups Tagging API
    # There are no warnings
    caplog.clear()
    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
        subnet_ids=[subnet_id],
        security_group_ids=[security_group_id],
    )
    assert len(client.list_services()) == 2

    assert len(client.list_services(tags={"foo": "bar"})) == 1
    assert client.list_services(tags={"foo": "bar"})[0].name == "tagged"
    assert len(caplog.records) == 0

    # Given that the Resource Group API returns resources that are not corresponding to services that
    # are present in the cluster, list_services should filter these out from its return value
    with mock.patch.object(client.tags_client, "get_paginator") as mock_get_paginator:
        mock_get_paginator.return_value.paginate.return_value = [
            {
                "ResourceTagMappingList": [
                    {
                        "ResourceARN": "arn:aws:ecs:us-east-1:123456789012:service/test/tagged",
                        "Tags": [{"Key": "foo", "Value": "bar"}],
                    },
                    {
                        "ResourceARN": "arn:aws:ecs:us-east-1:123456789012:service/test/missing",
                        "Tags": [{"Key": "foo", "Value": "bar"}],
                    },
                ]
            }
        ]
        result = client.list_services(tags={"foo": "bar"})
        assert len(result) == 1
        assert result[0].name == "tagged"


def test_delete_service_golden_path(aws_mock):
    """Test complete usual case where a service exist and is active, with a service discovery registration."""
    # Create VPC for service discovery namespace (moto 5 requires a real VPC)
    ec2 = boto3.client("ec2")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    vpc_id = vpc["Vpc"]["VpcId"]  # type: ignore

    ecs = boto3.client("ecs")
    service_discovery = boto3.client("servicediscovery")
    service_discovery.create_private_dns_namespace(Name="fake-namespace", Vpc=vpc_id)
    namespace = "fake-namespace"
    namespaces = service_discovery.list_namespaces(
        Filters=[{"Name": "NAME", "Values": [namespace]}]
    )
    namespace_id = namespaces["Namespaces"][0]["Id"]
    service_discovery.get_namespace(Id=namespace_id)
    ecs.create_cluster(clusterName="test", serviceConnectDefaults={"namespace": namespace_id})

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id=namespace_id,
        log_group="fake-log-group",
    )

    ecs.register_task_definition(
        family="test",
        executionRoleArn="test",
        taskRoleArn="test",
        containerDefinitions=[
            {
                "name": "test",
                "image": "test",
                "command": ["test"],
                "memory": 512,
            }
        ],
    )

    ecs.create_service(
        cluster="test",
        serviceName="my_service",
        taskDefinition="test",
    )

    service_discovery.create_service(Name="my_service", NamespaceId=namespace_id, Type="HTTP")

    # Moto does not implement or does not implement correctly these methods:
    # - service_discovery.list_instances
    # - service_discovery.deregister_instance
    # - service_discovery.get_operation
    # So we have to mock them manually, also we have to mock delete_service because it does not seem
    # to correctly change the list of services in the cluster.
    with mock.patch.object(
        client.service_discovery,
        "list_instances",
        return_value={"Instances": [{"Id": "fake_instance"}]},
    ):
        with mock.patch.object(
            client.service_discovery, "deregister_instance", return_value={"OperationId": "foobar"}
        ) as mock_deregister_instance:
            with mock.patch.object(
                client.service_discovery,
                "get_operation",
                return_value={"Operation": {"Status": "SUCCESS"}},
            ) as mock_get_operation:
                with mock.patch.object(client.ecs, "delete_service") as mock_delete_service:
                    StubService = namedtuple("Service", "name hostname")
                    client.delete_service(
                        StubService(name="my_service", hostname=f"my_service.{namespace}"),
                    )
                    assert mock_delete_service.call_args.kwargs == {
                        "cluster": "test",
                        "service": "my_service",
                        "force": True,
                    }
                assert mock_get_operation.call_args.kwargs == {"OperationId": "foobar"}
            assert mock_deregister_instance.call_args.kwargs["InstanceId"] == "fake_instance"


def test_delete_service_service_missing(aws_mock, caplog):
    """Test that the delete_service method does not raise an exception if the service is not found."""
    caplog.set_level("WARNING")
    ecs = boto3.client("ecs")
    ecs.create_cluster(clusterName="test")
    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="some-namespace-id",
        log_group="fake-log-group",
    )

    # moto's ServiceNotFoundException is not a subclass of botocore.exceptions.ClientError
    exception = botocore.exceptions.ClientError(  # pyright: ignore[reportAttributeAccessIssue]
        {"Error": {"Code": "ServiceNotFoundException", "Message": "Service was not found"}},
        "update_service",
    )

    StubService = namedtuple("Service", "name hostname")
    with mock.patch.object(client.ecs, "update_service", side_effect=exception):
        with mock.patch.object(client.ecs, "delete_service") as mock_delete_service:
            client.delete_service(
                StubService(name="my_service", hostname="my_service.some-namespace")
            )
            # delete will not have been called if the service is not found on the previous update command
            assert mock_delete_service.call_count == 0

    # command should not fail but should log a warning
    assert "my_service is not found or inactive" in caplog.text


def test_assign_public_ip_override():
    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
        assign_public_ip=True,
    )
    assert client.network_configuration["awsvpcConfiguration"]["assignPublicIp"] == "ENABLED"

    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
        assign_public_ip=False,
    )
    assert client.network_configuration["awsvpcConfiguration"]["assignPublicIp"] == "DISABLED"


def _make_sts_credentials(expires_in_minutes=60):
    """Create fake STS assume_role response credentials."""
    expiration = datetime.now(tz=timezone.utc) + timedelta(minutes=expires_in_minutes)
    return {
        "Credentials": {
            "AccessKeyId": "fake-access-key",
            "SecretAccessKey": "fake-secret-key",
            "SessionToken": "fake-session-token",
            "Expiration": expiration,
        },
        "AssumedRoleUser": {
            "AssumedRoleId": "fake-role-id",
            "Arn": "arn:aws:sts::123456789012:assumed-role/test-role/session",
        },
    }


def test_client_without_service_discovery_role_arn():
    """Client without service_discovery_role_arn uses default boto3 client."""
    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )
    assert client._service_discovery_role_arn is None
    assert client._service_discovery is not None


def test_client_with_service_discovery_role_arn():
    """Client with service_discovery_role_arn assumes role via STS."""
    role_arn = "arn:aws:iam::123456789012:role/cross-account-sd"
    fake_creds = _make_sts_credentials()

    with (
        mock.patch("boto3.client") as mock_boto_client,
        mock.patch("boto3.Session") as mock_session_cls,
    ):
        mock_sts = mock.MagicMock()
        mock_sts.assume_role.return_value = fake_creds

        def _client_factory(service, **kwargs):
            if service == "sts":
                return mock_sts
            return mock.MagicMock()

        mock_boto_client.side_effect = _client_factory
        mock_session = mock.MagicMock()
        mock_session_cls.return_value = mock_session

        client = Client(
            cluster_name="test",
            service_discovery_namespace_id="fake-namespace",
            log_group="fake-log-group",
            grace_period=0,
            service_discovery_role_arn=role_arn,
        )

        mock_sts.assume_role.assert_called_once()
        call_kwargs = mock_sts.assume_role.call_args.kwargs
        assert call_kwargs["RoleArn"] == role_arn
        assert call_kwargs["DurationSeconds"] == 3600
        assert "dagster-ecs-sd-" in call_kwargs["RoleSessionName"]

        mock_session_cls.assert_called_once_with(
            aws_access_key_id="fake-access-key",
            aws_secret_access_key="fake-secret-key",
            aws_session_token="fake-session-token",
        )

        mock_session.client.assert_called_once()
        assert client._service_discovery_role_arn == role_arn


def test_service_discovery_session_refresh():
    """Service discovery session is refreshed when near expiration."""
    role_arn = "arn:aws:iam::123456789012:role/cross-account-sd"

    with mock.patch("boto3.client") as mock_boto_client, mock.patch("boto3.Session"):
        mock_sts = mock.MagicMock()
        mock_sts.assume_role.return_value = _make_sts_credentials()

        def _client_factory(service, **kwargs):
            if service == "sts":
                return mock_sts
            return mock.MagicMock()

        mock_boto_client.side_effect = _client_factory

        client = Client(
            cluster_name="test",
            service_discovery_namespace_id="fake-namespace",
            log_group="fake-log-group",
            grace_period=0,
            service_discovery_role_arn=role_arn,
        )

        assert mock_sts.assume_role.call_count == 1

        # Access the property — session is still valid, no refresh
        _ = client.service_discovery
        assert mock_sts.assume_role.call_count == 1

        # Simulate session about to expire (within 5-minute buffer)
        client._sd_session_expires = datetime.now(tz=timezone.utc) + timedelta(minutes=2)
        mock_sts.assume_role.return_value = _make_sts_credentials()

        # Access triggers refresh
        _ = client.service_discovery
        assert mock_sts.assume_role.call_count == 2


def test_service_discovery_no_refresh_without_role_arn():
    """Without role ARN, service_discovery property returns client directly without refresh."""
    client = Client(
        cluster_name="test",
        service_discovery_namespace_id="fake-namespace",
        log_group="fake-log-group",
        grace_period=0,
    )

    sd_client = client.service_discovery
    # Accessing multiple times returns the same client, no refresh logic involved
    assert client.service_discovery is sd_client



import datetime
import pickle
from unittest import mock

import boto3
import pytest
from dagster import (
    GraphIn,
    GraphOut,
    In,
    StaticPartitionsDefinition,
    TimeWindowPartitionMapping,
    asset,
    graph,
    job,
    materialize,
    op,
)
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.unresolved_asset_job_definition import define_asset_job
from dagster._core.test_utils import environ
from dagster._time import get_current_datetime
from dagster_cloud.serverless.io_manager import (
    PickledObjectServerlessIOManager,
    serverless_io_manager,
)
from mypy_boto3_s3.service_resource import S3ServiceResource


@pytest.fixture
def mock_s3_bucket(s3_resource: S3ServiceResource):
    bucket = s3_resource.create_bucket(Bucket="test-bucket")
    try:
        yield bucket
    finally:
        bucket.objects.all().delete()
        bucket.delete()


@pytest.fixture
def mock_boto_session(s3_resource):
    yield boto3.Session()


@pytest.fixture
def serverless_io_manager_env():
    with environ(
        {
            "DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET": "test-bucket",
            "DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX": "test-prefix",
        },
    ):
        yield


@pytest.fixture
def mock_refresh_boto_session(mock_boto_session):
    with mock.patch.object(
        PickledObjectServerlessIOManager,
        "_refresh_boto_session",
        return_value=(mock_boto_session, get_current_datetime() + datetime.timedelta(days=3)),
    ):
        yield


def get_inty_job():
    @op
    def first_op(first_input):
        assert first_input == 4
        return first_input * 2

    @op
    def second_op(second_input):
        assert second_input == 8
        return second_input + 3

    source1 = SourceAsset("source1", partitions_def=StaticPartitionsDefinition(["foo", "bar"]))

    @asset
    def asset1(source1):
        return source1["foo"] + source1["bar"]

    @asset
    def asset2(asset1):
        assert asset1 == 3
        return asset1 + 1

    @graph(ins={"asset2": GraphIn()}, out={"asset3": GraphOut()})
    def graph_asset(asset2):
        return second_op(first_op(asset2))

    @asset(partitions_def=StaticPartitionsDefinition(["apple", "orange"]))
    def partitioned():
        return 8

    graph_asset_def = AssetsDefinition.from_graph(graph_asset)
    target_assets = [asset1, asset2, graph_asset_def, partitioned]

    return Definitions(
        assets=[*target_assets, source1],
        jobs=[define_asset_job("assets", target_assets)],
        resources={"io_manager": serverless_io_manager},
    ).resolve_job_def("assets")


@op
def int_op():
    return 1


@op(ins={"int_in": In(int)})
def io_manager_op(context, int_in):
    from dagster_cloud.serverless.io_manager import PickledObjectServerlessIOManager

    assert int_in == 1
    assert type(context.resources.io_manager) == PickledObjectServerlessIOManager
    return 2


@job(resource_defs={"io_manager": serverless_io_manager})
def serverless_io_manager_job():
    io_manager_op(int_in=int_op())


def test_serverless_io_manager_job_execution(
    mock_s3_bucket,
    serverless_io_manager_env,
    agent_instance_local_ursula,
    mock_refresh_boto_session,
):
    instance = agent_instance_local_ursula
    result = serverless_io_manager_job.execute_in_process(instance=instance)
    assert result.success
    assert result.output_for_node("int_op") == 1
    assert result.output_for_node("io_manager_op") == 2


from dagster import AssetExecutionContext, AssetIn, DailyPartitionsDefinition


def test_serverless_io_manager_allow_missing_partitions(
    mock_s3_bucket,
    serverless_io_manager_env,
    agent_instance_local_ursula,
    mock_refresh_boto_session,
):
    start = datetime.datetime(2022, 1, 1)

    daily = DailyPartitionsDefinition(start_date=f"{start:%Y-%m-%d}")

    @asset(partitions_def=daily, io_manager_def=serverless_io_manager)
    def upstream_asset(context: AssetExecutionContext) -> str:
        return context.partition_key

    @asset(
        partitions_def=daily,
        io_manager_def=serverless_io_manager,
        ins={
            "upstream_asset": AssetIn(
                partition_mapping=TimeWindowPartitionMapping(start_offset=-1),
                metadata={"allow_missing_partitions": True},
            )
        },
    )
    def downstream_asset(upstream_asset: dict[str, str]):
        return upstream_asset

    materialize(
        [upstream_asset],
        partition_key=start.strftime(daily.fmt),
        instance=agent_instance_local_ursula,
    )
    result = materialize(
        [upstream_asset.to_source_asset(), downstream_asset],
        partition_key=(start + datetime.timedelta(days=1)).strftime(daily.fmt),
        instance=agent_instance_local_ursula,
    )
    downstream_asset_data = result.output_for_node("downstream_asset", "result")
    assert len(downstream_asset_data) == 1, "1 partition should be missing"


def test_serverless_io_manager_asset_execution(
    mock_s3_bucket,
    serverless_io_manager_env,
    agent_instance_local_ursula,
    mock_refresh_boto_session,
):
    instance = agent_instance_local_ursula

    inty_job = get_inty_job()

    assert not len(list(mock_s3_bucket.objects.all()))

    serverless_prefix = "test-prefix/io_storage/sandbox"

    mock_s3_bucket.put_object(Key=f"{serverless_prefix}/source1/foo", Body=pickle.dumps(1))
    # pickled_source1_bar = pickle.dumps(2)
    mock_s3_bucket.put_object(Key=f"{serverless_prefix}/source1/bar", Body=pickle.dumps(2))

    result = inty_job.execute_in_process(partition_key="apple", instance=instance)

    assert result.output_for_node("asset1") == 3
    assert result.output_for_node("asset2") == 4
    assert result.output_for_node("graph_asset.first_op") == 8
    assert result.output_for_node("graph_asset.second_op") == 11

    objects = list(mock_s3_bucket.objects.all())
    assert len(objects) == 7
    assert {(o.bucket_name, o.key) for o in objects} == {
        ("test-bucket", f"{serverless_prefix}/source1/bar"),
        ("test-bucket", f"{serverless_prefix}/source1/foo"),
        ("test-bucket", f"{serverless_prefix}/asset1"),
        ("test-bucket", f"{serverless_prefix}/asset2"),
        ("test-bucket", f"{serverless_prefix}/asset3"),
        ("test-bucket", f"{serverless_prefix}/partitioned/apple"),
        (
            "test-bucket",
            "/".join(
                [serverless_prefix, "storage", result.run_id, "graph_asset.first_op", "result"]
            ),
        ),
    }

    # re-execution does not cause issues, overwrites the buckets
    result2 = inty_job.execute_in_process(partition_key="apple", instance=instance)

    objects = list(mock_s3_bucket.objects.all())
    assert len(objects) == 8
    assert {(o.bucket_name, o.key) for o in objects} == {
        ("test-bucket", f"{serverless_prefix}/source1/bar"),
        ("test-bucket", f"{serverless_prefix}/source1/foo"),
        ("test-bucket", f"{serverless_prefix}/asset1"),
        ("test-bucket", f"{serverless_prefix}/asset2"),
        ("test-bucket", f"{serverless_prefix}/asset3"),
        ("test-bucket", f"{serverless_prefix}/partitioned/apple"),
        (
            "test-bucket",
            "/".join(
                [serverless_prefix, "storage", result.run_id, "graph_asset.first_op", "result"]
            ),
        ),
        (
            "test-bucket",
            "/".join(
                [
                    serverless_prefix,
                    "storage",
                    result2.run_id,
                    "graph_asset.first_op",
                    "result",
                ]
            ),
        ),
    }



from dagster import In, Int, Out, job, op
from dagster_cloud.serverless.io_manager import serverless_io_manager


@op(out=Out(Int))
def return_one():
    return 1


@op(
    ins={"num": In(Int)},
    out=Out(Int),
)
def add_one(num):
    return num + 1


@job(
    resource_defs={
        "io_manager": serverless_io_manager,
    }
)
def serverless_job():
    add_one(return_one())


def test_set_explicitly():
    # test that we can import and set the serverless_io_manager explicitly if desired
    assert serverless_job



import tempfile
import time
from contextlib import contextmanager
from unittest import mock

import boto3
import moto
import pytest
from dagster import DagsterRunStatus
from dagster._config import process_config, resolve_to_config_type
from dagster._core.launcher.base import WorkerStatus
from dagster._core.storage.tags import RUN_WORKER_ID_TAG
from dagster._core.test_utils import create_run_for_test, environ, instance_for_test
from dagster._serdes import ConfigurableClassData
from dagster._utils.merger import merge_dicts
from dagster_aws.ecs import EcsRunLauncher
from dagster_cloud.execution.monitoring import CloudRunWorkerStatus
from dagster_cloud.workspace.ecs import EcsUserCodeLauncher
from dagster_cloud.workspace.ecs.client import (
    DEFAULT_ECS_GRACE_PERIOD,
    DEFAULT_ECS_TIMEOUT,
    Service,
)
from dagster_cloud.workspace.user_code_launcher import UserCodeLauncherEntry
from dagster_cloud.workspace.user_code_launcher.utils import (
    deterministic_label_for_location,
    get_grpc_server_env,
)
from dagster_cloud_cli.core.workspace import CodeLocationDeployData


@pytest.fixture
def secrets_manager():
    with moto.mock_aws():
        yield boto3.client("secretsmanager")


@contextmanager
def ecs_instance(monkeypatch, user_code_launcher_overrides=None, overrides=None):
    monkeypatch.setenv("CLUSTER", "fake-cluster")
    monkeypatch.setenv("SUBNET1", "fake-subnet-1")
    monkeypatch.setenv("SUBNET2", "fake-subnet-2")

    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.ecs",
                "class": "EcsUserCodeLauncher",
                "config": merge_dicts(
                    {
                        "cluster": {
                            "env": "CLUSTER",
                        },
                        "subnets": [
                            {"env": "SUBNET1"},
                            {"env": "SUBNET2"},
                        ],
                        "service_discovery_namespace_id": "fake-namespace",
                        "execution_role_arn": "fake-role",
                        "log_group": "fake-log-group",
                    },
                    user_code_launcher_overrides or {},
                ),
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": "FAKE_TOKEN",
                "deployment": "sandbox",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
            **(overrides or {}),
        }
    ) as instance:
        yield instance


@moto.mock_aws()
def test_default_ecs_instance(monkeypatch):
    with ecs_instance(monkeypatch) as instance:
        assert isinstance(instance.user_code_launcher, EcsUserCodeLauncher)  # pyright: ignore[reportAttributeAccessIssue]
        assert isinstance(instance.user_code_launcher.inst_data, ConfigurableClassData)  # pyright: ignore[reportAttributeAccessIssue]
        assert "fake-subnet-1" in instance.user_code_launcher.subnets  # pyright: ignore[reportAttributeAccessIssue]
        assert "fake-subnet-2" in instance.user_code_launcher.subnets  # pyright: ignore[reportAttributeAccessIssue]
        assert len(instance.user_code_launcher.subnets) == 2  # pyright: ignore[reportAttributeAccessIssue]
        assert not instance.user_code_launcher.task_role_arn  # pyright: ignore[reportAttributeAccessIssue]
        assert not instance.user_code_launcher.security_group_ids  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.secrets == []  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.secrets_tag is None  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.env_vars == []  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.client.timeout == DEFAULT_ECS_TIMEOUT  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.client.grace_period == DEFAULT_ECS_GRACE_PERIOD  # pyright: ignore[reportAttributeAccessIssue]

        assert not instance.user_code_launcher._get_enable_ecs_exec()  # noqa  # pyright: ignore[reportAttributeAccessIssue]


@moto.mock_aws()
def test_timeouts_override(monkeypatch):
    with ecs_instance(monkeypatch, {"ecs_timeout": 100, "ecs_grace_period": 50}) as instance:
        assert instance.user_code_launcher.client.timeout == 100  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.client.grace_period == 50  # pyright: ignore[reportAttributeAccessIssue]


@moto.mock_aws()
def test_ecs_exec_override(monkeypatch):
    with ecs_instance(monkeypatch, {"enable_ecs_exec": True}) as instance:
        assert instance.user_code_launcher._get_enable_ecs_exec()  # noqa  # pyright: ignore[reportAttributeAccessIssue]


@moto.mock_aws()
def test_env_vars_override(monkeypatch):
    env_vars = ["FOO_ENV_VAR", "BAR_ENV_VAR=BAR_VALUE"]
    with ecs_instance(monkeypatch, {"env_vars": env_vars}) as instance:
        assert instance.user_code_launcher.env_vars == env_vars  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.run_launcher.env_vars == env_vars  # pyright: ignore[reportAttributeAccessIssue]


@moto.mock_aws()
def test_secrets_override(monkeypatch):
    secrets = [
        {
            "name": "FOO",
            "valueFrom": "arn:aws:secretsmanager:us-west-2:111122223333:secret:aes128-1a2b3c",
        },
        {
            "name": "BAR",
            "valueFrom": "arn:aws:secretsmanager:us-west-2:111122223333:secret:aes192-4D5e6F",
        },
    ]
    with ecs_instance(monkeypatch, {"secrets": secrets, "secrets_tag": "my_tag"}) as instance:
        assert instance.user_code_launcher.secrets == secrets  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.secrets_tag == "my_tag"  # pyright: ignore[reportAttributeAccessIssue]
        assert isinstance(instance.user_code_launcher.inst_data, ConfigurableClassData)  # pyright: ignore[reportAttributeAccessIssue]

        assert instance.run_launcher.secrets == secrets  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.run_launcher.secrets_tags == ["my_tag"]  # pyright: ignore[reportAttributeAccessIssue]

    with ecs_instance(monkeypatch, {"task_role_arn": "fake-role"}) as instance:
        assert instance.user_code_launcher.task_role_arn == "fake-role"  # pyright: ignore[reportAttributeAccessIssue]

    with ecs_instance(monkeypatch, {"security_group_ids": ["fake-sg"]}) as instance:
        assert instance.user_code_launcher.security_group_ids == ["fake-sg"]  # pyright: ignore[reportAttributeAccessIssue]


@moto.mock_aws()
def test_empty_secrets(monkeypatch):
    with ecs_instance(monkeypatch, {"secrets": [], "secrets_tag": None}) as instance:
        assert instance.user_code_launcher.secrets == []  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.secrets_tag is None  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.run_launcher.secrets == []  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.run_launcher.secrets_tags == []  # pyright: ignore[reportAttributeAccessIssue]


@moto.mock_aws()
def test_timeout_debug_info(monkeypatch):
    with ecs_instance(monkeypatch) as instance:
        fake_client = mock.MagicMock()
        fake_client.get_task_logs = mock.MagicMock(return_value=["Hi", "these", "are", "logs"])

        instance.user_code_launcher.client = fake_client  # pyright: ignore[reportAttributeAccessIssue]

        timeout_debug_info = instance.user_code_launcher._get_update_failure_debug_info(  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            "my_task_arn"
        )

        assert (
            timeout_debug_info
            == "Task logs:\nHi\nthese\nare\nlogs\n\nFor more information about the failure, check"
            " the ECS console for logs for task my_task_arn in cluster fake-cluster."
        )


@moto.mock_aws()
def test_transient_startup_failure_health_check(monkeypatch):
    with tempfile.TemporaryDirectory() as temp_dir:
        with ecs_instance(
            monkeypatch,
            overrides={
                "run_storage": {
                    "module": "dagster._core.storage.runs",
                    "class": "SqliteRunStorage",
                    "config": {"base_dir": temp_dir},
                },
            },
        ) as instance:
            fake_ecs = mock.MagicMock()

            task_arn = "arn:aws:ecs:us-west-2:657821118200:task/serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster/d12fcac0355f4bb3ae768653d61cc31e"

            # taken from a real transient startup failure with some irrelevant keys stripped out
            fake_ecs.describe_tasks = mock.MagicMock(
                return_value={
                    "tasks": [
                        {
                            "containers": [
                                {
                                    "containerArn": "arn:aws:ecs:us-west-2:657821118200:container/serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster/d12fcac0355f4bb3ae768653d61cc31e/8d008fad-e39f-4fd6-a4f5-324a7c2f0eff",
                                    "taskArn": "arn:aws:ecs:us-west-2:657821118200:task/serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster/d12fcac0355f4bb3ae768653d61cc31e",
                                    "name": "DatadogAgent",
                                    "image": "public.ecr.aws/datadog/agent:latest",
                                    "lastStatus": "STOPPED",
                                    "healthStatus": "UNKNOWN",
                                    "cpu": "0",
                                },
                                {
                                    "containerArn": "arn:aws:ecs:us-west-2:657821118200:container/serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster/d12fcac0355f4bb3ae768653d61cc31e/c40d570e-0243-4bde-a760-e965ebf938a8",
                                    "taskArn": task_arn,
                                    "name": "dagster",
                                    "image": "657821118200.dkr.ecr.us-west-2.amazonaws.com/fake-image-29f7eac-5070495852-1",
                                    "lastStatus": "STOPPED",
                                    "healthStatus": "UNKNOWN",
                                    "cpu": "0",
                                },
                            ],
                            "cpu": "4096",
                            "desiredStatus": "STOPPED",
                            "enableExecuteCommand": False,
                            "group": "family:fake_task_definition_9972342a",
                            "healthStatus": "UNKNOWN",
                            "lastStatus": "STOPPED",
                            "launchType": "FARGATE",
                            "platformVersion": "1.4.0",
                            "platformFamily": "Linux",
                            "stopCode": "TaskFailedToStart",
                            "stoppedReason": (
                                "Timeout waiting for network interface provisioning to complete."
                            ),
                            "tags": [],
                            "taskArn": "arn:aws:ecs:us-west-2:657821118200:task/serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster/d12fcac0355f4bb3ae768653d61cc31e",
                            "taskDefinitionArn": "arn:aws:ecs:us-west-2:657821118200:task-definition/fun-fake-group-dagster-cloud_9972342a:1966",
                            "version": 4,
                            "ephemeralStorage": {"sizeInGiB": 128},
                        }
                    ],
                    "failures": [],
                }
            )

            run_launcher = instance.run_launcher
            run_launcher.ecs = fake_ecs  # pyright: ignore[reportAttributeAccessIssue]

            starting_run = create_run_for_test(
                instance,
                job_name="foo_job",
                tags={
                    RUN_WORKER_ID_TAG: "abcde",
                    "ecs/task_arn": task_arn,
                    "ecs/cluster": "serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster",
                },
                status=DagsterRunStatus.STARTING,
            )

            health_check = CloudRunWorkerStatus.from_check_run_health_result(
                starting_run.run_id, run_launcher.check_run_worker_health(starting_run)
            )
            assert health_check.status_type == WorkerStatus.FAILED
            assert health_check.transient
            assert health_check.run_worker_id == "abcde"

            # started run is not considered transient
            started_run = create_run_for_test(
                instance,
                job_name="foo_job",
                tags={
                    RUN_WORKER_ID_TAG: "abcde",
                    "ecs/task_arn": task_arn,
                    "ecs/cluster": "serverless-agent-5bae4d6a-b58f-392d-b7ab-92c53d938193-Cluster",
                },
                status=DagsterRunStatus.STARTED,
            )

            health_check = CloudRunWorkerStatus.from_check_run_health_result(
                started_run.run_id, run_launcher.check_run_worker_health(started_run)
            )
            assert not health_check.transient


@moto.mock_aws()
def test_cannot_set_system_tags(monkeypatch):
    with ecs_instance(
        monkeypatch,
        {
            "server_ecs_tags": [
                {
                    "key": "dagster/deployment_name",
                    "value": "NOPE",
                },
                {
                    "key": "dagster/grpc_server",
                },
            ],
        },
    ) as instance:
        instance.user_code_launcher._wait_for_dagster_server_process = mock.MagicMock(  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            return_value=None
        )  # fmt: skip

        metadata_with_container_context = CodeLocationDeployData(
            image="foo_image",
            python_file="repo.py",
        )

        with pytest.raises(Exception, match="Cannot override system ECS tags"):
            instance.user_code_launcher._start_new_server_spinup(  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                deployment_name="sandbox",
                location_name="test_location",
                desired_entry=UserCodeLauncherEntry(metadata_with_container_context, time.time()),
            )


@moto.mock_aws()
def test_config_from_container_context(monkeypatch, secrets_manager):
    foo_tagged_secret = secrets_manager.create_secret(
        Name="foo_tagged_secret",
        SecretString="foo_tagged_value",
        Tags=[{"Key": "my_tag", "Value": "foo"}],
    )

    bar_tagged_secret = secrets_manager.create_secret(
        Name="bar_tagged_secret",
        SecretString="bar_tagged_value",
        Tags=[{"Key": "additional_tag", "Value": "bar"}],
    )

    foo_secret_arn = "arn:aws:secretsmanager:us-west-2:111122223333:secret:aes128-1a2b3c"

    hello_secret_arn = "arn:aws:secretsmanager:us-east-1:123456789012:secret:HELLO-AbCdEf:token::"

    instance_secrets = [
        {"name": "FOO", "valueFrom": foo_secret_arn},
    ]
    instance_env_vars = ["FOO_ENV_VAR"]
    with ecs_instance(
        monkeypatch,
        {
            "secrets": instance_secrets,
            "secrets_tag": "my_tag",
            "env_vars": instance_env_vars,
            "runtime_platform": {"operatingSystemFamily": "LINUX"},
            "mount_points": [
                {
                    "sourceVolume": "myEfsVolume",
                    "containerPath": "/mount/efs",
                    "readOnly": True,
                }
            ],
            "volumes": [
                {
                    "name": "myEfsVolume",
                    "efsVolumeConfiguration": {
                        "fileSystemId": "fs-1234",
                        "rootDirectory": "/path/to/my/data",
                    },
                }
            ],
            "server_sidecar_containers": [
                {
                    "name": "DatadogAgent",
                    "image": "public.ecr.aws/datadog/agent:latest",
                    "environment": [
                        {"name": "ECS_FARGATE", "value": "true"},
                    ],
                }
            ],
            "run_sidecar_containers": [
                {
                    "name": "busyrun",
                    "image": "busybox:latest",
                }
            ],
            "server_ecs_tags": [
                {
                    "key": "FOO",  # no value
                }
            ],
            "run_ecs_tags": [{"key": "ABC", "value": "DEF"}],  # with value
            "server_health_check": {
                "command": ["HELLO"],
            },
        },
    ) as instance:
        fake_client = mock.MagicMock()
        fake_client.create_service = mock.MagicMock(
            return_value=Service(
                client=fake_client,
                arn="arn:aws:ecs::us-west-2:123456:service/my-cluster-name/my-service-name",
            )
        )

        instance.user_code_launcher.client = fake_client  # pyright: ignore[reportAttributeAccessIssue]

        instance.user_code_launcher._wait_for_dagster_server_process = mock.MagicMock(  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            return_value=None
        )  # fmt: skip

        metadata_with_container_context = CodeLocationDeployData(
            image="foo_image",
            python_file="repo.py",
            container_context={
                "ecs": {
                    "secrets": [
                        {
                            "name": "HELLO",
                            "valueFrom": hello_secret_arn,
                        }
                    ],
                    "secrets_tags": ["additional_tag"],
                    "env_vars": ["BAR_ENV_VAR=BAR_VALUE"],
                    "server_resources": {
                        "cpu": "1024",
                        "memory": "2048",
                    },
                    "task_role_arn": "fake-task-role-arn",
                    "execution_role_arn": "fake-execution-role-arn",
                    "runtime_platform": {"operatingSystemFamily": "WINDOWS_SERVER_2019_FULL"},
                    "mount_points": [
                        {
                            "sourceVolume": "myOtherEfsVolume",
                            "containerPath": "/mount/other/efs",
                            "readOnly": True,
                        }
                    ],
                    "volumes": [
                        {
                            "name": "myOtherEfsVolume",
                            "efsVolumeConfiguration": {
                                "fileSystemId": "fs-5678",
                                "rootDirectory": "/path/to/my/other/data",
                            },
                        },
                    ],
                    "server_sidecar_containers": [
                        {
                            "name": "OtherServerAgent",
                            "image": "public.ecr.aws/other/agent:latest",
                        }
                    ],
                    "run_sidecar_containers": [
                        {
                            "name": "OtherRunAgent",
                            "image": "otherrun:latest",
                        }
                    ],
                    "server_ecs_tags": [{"key": "BAZ", "value": "QUUX"}],
                    "run_ecs_tags": [
                        {
                            "key": "GHI",
                        }
                    ],
                    "repository_credentials": "fake-secret-arn",
                    "server_health_check": {
                        "command": ["CMD-SHELL", "curl -f http://localhost/ || exit 1"],
                        "interval": 30,
                        "timeout": 5,
                        "retries": 3,
                        "startPeriod": 0,
                    },
                }
            },
        )

        server_timestamp = time.time()

        with environ({"FOO_ENV_VAR": "FOO_VALUE"}):
            instance.user_code_launcher._start_new_server_spinup(  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
                deployment_name="sandbox",
                location_name="test_location",
                desired_entry=UserCodeLauncherEntry(
                    metadata_with_container_context, server_timestamp
                ),
            )

        instance.user_code_launcher.client.create_service.assert_called_once()  # pyright: ignore[reportAttributeAccessIssue]
        _args, kwargs = instance.user_code_launcher.client.create_service.call_args  # pyright: ignore[reportAttributeAccessIssue]

        expected_kwargs = {
            "image": "foo_image",
            "command": metadata_with_container_context.get_grpc_server_command(),
            "env": {
                "FOO_ENV_VAR": "FOO_VALUE",
                "BAR_ENV_VAR": "BAR_VALUE",
                **get_grpc_server_env(
                    metadata_with_container_context,
                    4000,
                    location_name="test_location",
                    instance_ref=instance.ref_for_deployment(  # pyright: ignore[reportAttributeAccessIssue]
                        "sandbox",
                    ),
                ),
            },
            "tags": {
                "dagster/location_name": "testlocation",
                "dagster/deployment_name": "sandbox",
                "dagster/location_hash": deterministic_label_for_location(
                    "sandbox", "test_location"
                ),
                "dagster/grpc_server": "1",
                "dagster/agent_id": instance.instance_uuid,  # pyright: ignore[reportAttributeAccessIssue]
                "FOO": None,
                "BAZ": "QUUX",
                "dagster/server_timestamp": str(server_timestamp),
            },
            "task_role_arn": "fake-task-role-arn",
            "execution_role_arn": "fake-execution-role-arn",
            "secrets": {
                "FOO": foo_secret_arn,
                "HELLO": hello_secret_arn,
                "bar_tagged_secret": bar_tagged_secret["ARN"],
                "foo_tagged_secret": foo_tagged_secret["ARN"],
            },
            "cpu": "1024",
            "memory": "2048",
            "runtime_platform": {"operatingSystemFamily": "WINDOWS_SERVER_2019_FULL"},
            "mount_points": [
                {
                    "sourceVolume": "myOtherEfsVolume",
                    "containerPath": "/mount/other/efs",
                    "readOnly": True,
                },
                {
                    "sourceVolume": "myEfsVolume",
                    "containerPath": "/mount/efs",
                    "readOnly": True,
                },
            ],
            "volumes": [
                {
                    "name": "myOtherEfsVolume",
                    "efsVolumeConfiguration": {
                        "fileSystemId": "fs-5678",
                        "rootDirectory": "/path/to/my/other/data",
                    },
                },
                {
                    "name": "myEfsVolume",
                    "efsVolumeConfiguration": {
                        "fileSystemId": "fs-1234",
                        "rootDirectory": "/path/to/my/data",
                    },
                },
            ],
            "sidecars": [
                {
                    "name": "OtherServerAgent",
                    "image": "public.ecr.aws/other/agent:latest",
                },
                {
                    "name": "DatadogAgent",
                    "image": "public.ecr.aws/datadog/agent:latest",
                    "environment": [
                        {"name": "ECS_FARGATE", "value": "true"},
                    ],
                },
            ],
            "repository_credentials": "fake-secret-arn",
            "health_check": {
                "command": ["CMD-SHELL", "curl -f http://localhost/ || exit 1"],
                "interval": 30,
                "timeout": 5,
                "retries": 3,
                "startPeriod": 0,
            },
        }

        differing_keys = [key for key, value in expected_kwargs.items() if kwargs[key] != value]

        assert not differing_keys, (
            f"Differing keys: {differing_keys} between {kwargs} and {expected_kwargs}"
        )

        run_launcher_kwargs = instance.user_code_launcher._run_launcher_kwargs()  # noqa  # pyright: ignore[reportAttributeAccessIssue]

        process_result = process_config(
            resolve_to_config_type(EcsRunLauncher.config_type()),
            run_launcher_kwargs,
        )
        assert process_result.success, str(process_result.errors)

        assert run_launcher_kwargs == {
            "task_definition": {
                "log_group": "fake-log-group",
                "execution_role_arn": "fake-role",
                "requires_compatibilities": ["FARGATE"],
                "runtime_platform": {"operatingSystemFamily": "LINUX"},
                "mount_points": [
                    {
                        "sourceVolume": "myEfsVolume",
                        "containerPath": "/mount/efs",
                        "readOnly": True,
                    }
                ],
                "volumes": [
                    {
                        "name": "myEfsVolume",
                        "efsVolumeConfiguration": {
                            "fileSystemId": "fs-1234",
                            "rootDirectory": "/path/to/my/data",
                        },
                    }
                ],
                "sidecar_containers": [
                    {
                        "name": "busyrun",
                        "image": "busybox:latest",
                    }
                ],
            },
            "secrets": instance_secrets,
            "secrets_tag": "my_tag",
            "container_name": "dagster",
            "env_vars": ["FOO_ENV_VAR"],
            "use_current_ecs_task_config": False,
            "run_resources": {},
            "run_task_kwargs": {
                "cluster": "fake-cluster",
                "networkConfiguration": fake_client.network_configuration,
                "launchType": "FARGATE",
            },
            "run_ecs_tags": [
                {"key": "ABC", "value": "DEF"},
            ],
            "task_definition_prefix": "dagsterrun",
        }



import warnings
from collections.abc import Generator, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from io import StringIO
from typing import Any, cast

import snowflake.connector
from dagster import (
    AssetExecutionContext,
    AssetKey,
    DagsterInvariantViolationError,
    OpExecutionContext,
    get_dagster_logger,
)
from dagster._annotations import beta, public
from dagster._check import CheckError
from dagster._core.errors import DagsterInvalidPropertyError
from dagster._core.storage.event_log.sql_event_log import SqlDbConnection
from dagster_snowflake import SnowflakeConnection, SnowflakeResource
from snowflake.connector.cursor import SnowflakeCursor

from dagster_cloud.dagster_insights.insights_utils import get_current_context_and_asset_key
from dagster_cloud.dagster_insights.snowflake.snowflake_utils import meter_snowflake_query


def get_current_context_and_asset_key_or_warn() -> tuple[
    OpExecutionContext | AssetExecutionContext | None, AssetKey | None
]:
    try:
        return get_current_context_and_asset_key()
    except (DagsterInvalidPropertyError, DagsterInvariantViolationError, CheckError):
        warnings.warn(
            "Accessed InsightsSnowflakeResource outside of an Op or Asset context."
            " This query may not be properly attributed."
        )
        return None, None


class InsightsSnowflakeCursor(SnowflakeCursor):
    def __init__(self, *args, **kwargs) -> None:
        self._asset_key = None
        super().__init__(*args, **kwargs)

    def set_asset_key(self, asset_key: AssetKey | None):
        self._asset_key = asset_key

    def execute(self, command: str, *args, **kwargs):
        context, inferred_asset_key = get_current_context_and_asset_key_or_warn()
        if not context:
            return super().execute(command, *args, **kwargs)

        associated_asset_key = self._asset_key or inferred_asset_key

        return super().execute(
            meter_snowflake_query(context, command, associated_asset_key=associated_asset_key),
            *args,
            **kwargs,
        )


class WrappedSnowflakeConnection(snowflake.connector.SnowflakeConnection):
    def __init__(self, *args, asset_key: AssetKey | None = None, **kwargs) -> None:
        self._asset_key = asset_key
        super().__init__(*args, **kwargs)

    def execute_string(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        sql_text: str,
        remove_comments: bool = False,
        return_cursors: bool = True,
        cursor_class: type[SnowflakeCursor] = InsightsSnowflakeCursor,
        **kwargs,
    ) -> Iterable[SnowflakeCursor]:
        return super().execute_string(
            sql_text,
            remove_comments,
            return_cursors,
            cursor_class,  # type: ignore  # (bad stubs)
            **kwargs,
        )

    def execute_stream(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        stream: StringIO,
        remove_comments: bool = False,
        cursor_class: type[SnowflakeCursor] = InsightsSnowflakeCursor,
        **kwargs,
    ) -> Generator[SnowflakeCursor, None, None]:
        return super().execute_stream(stream, remove_comments, cursor_class, **kwargs)  # type: ignore  # (bad stubs)

    def cursor(self, cursor_class=None) -> SnowflakeCursor:
        if cursor_class is None:
            cursor = cast("InsightsSnowflakeCursor", super().cursor(InsightsSnowflakeCursor))
            cursor.set_asset_key(self._asset_key)
        else:
            cursor = super().cursor(cursor_class)

        return cursor


@beta
class InsightsSnowflakeResource(SnowflakeResource):
    """A wrapper around :py:class:`SnowflakeResource` which automatically tags
    Snowflake queries with comments which can be used to attribute Snowflake
    query costs to Dagster jobs and assets.

    If connector configuration is not set, InsightsSnowflakeResource.get_connection() will return a
    `snowflake.connector.Connection <https://docs.snowflake.com/en/developer-guide/python-connector/python-connector-api#object-connection>`__
    object. If connector="sqlalchemy" configuration is set, then InsightsSnowflakeResource.get_connection() will
    return a `SQLAlchemy Connection <https://docs.sqlalchemy.org/en/20/core/connections.html#sqlalchemy.engine.Connection>`__
    or a `SQLAlchemy raw connection <https://docs.sqlalchemy.org/en/20/core/connections.html#sqlalchemy.engine.Engine.raw_connection>`__.

    A simple example of loading data into Snowflake and subsequently querying that data is shown below:

    Examples:
        .. code-block:: python

            from dagster import job, op
            from dagster_insights import InsightsSnowflakeResource

            @op
            def get_one(snowflake_resource: InsightsSnowflakeResource):
                with snowflake_resource.get_connection() as conn:
                    # conn is a snowflake.connector.Connection object
                    conn.cursor().execute("SELECT 1")

            @job
            def my_snowflake_job():
                get_one()

            my_snowflake_job.execute_in_process(
                resources={
                    'snowflake_resource': InsightsSnowflakeResource(
                        account=EnvVar("SNOWFLAKE_ACCOUNT"),
                        user=EnvVar("SNOWFLAKE_USER"),
                        password=EnvVar("SNOWFLAKE_PASSWORD")
                        database="MY_DATABASE",
                        schema="MY_SCHEMA",
                        warehouse="MY_WAREHOUSE"
                    )
                }
            )
    """

    def get_object_to_set_on_execution_context(self) -> Any:
        # Directly create a SnowflakeConnection here for backcompat since the SnowflakeConnection
        # has methods this resource does not have
        return SnowflakeConnection(
            config=self._resolved_config_dict,
            log=get_dagster_logger(),
            snowflake_connection_resource=self,
        )

    @public
    @contextmanager
    def get_connection(
        self, raw_conn: bool = True
    ) -> Iterator[SqlDbConnection | WrappedSnowflakeConnection]:
        if self.connector == "sqlalchemy":
            from snowflake.sqlalchemy import URL
            from sqlalchemy import create_engine, event

            engine = create_engine(
                URL(**self._sqlalchemy_connection_args),
                connect_args=self._sqlalchemy_engine_args,
            )

            # Attach a listener to the connection which will add a comment to any SQL query
            # executed through the connection. This comment will be used to identify the query
            # when later attributing cost.
            @event.listens_for(engine, "before_cursor_execute", retval=True)
            def comment_sql(conn, cursor, statement, parameters, context, executemany):
                context, asset_key = get_current_context_and_asset_key_or_warn()
                if not context:
                    return statement, parameters

                statement = meter_snowflake_query(
                    context, statement, associated_asset_key=asset_key
                )
                return statement, parameters

            conn = engine.raw_connection() if raw_conn else engine.connect()

            yield conn
            conn.close()
            engine.dispose()
        else:
            conn = WrappedSnowflakeConnection(**self._connection_args)

            yield conn
            if not self.autocommit:
                conn.commit()
            conn.close()

    @public
    @contextmanager
    def get_connection_for_asset(
        self, asset_key: AssetKey, raw_conn: bool = True
    ) -> Iterator[SqlDbConnection | WrappedSnowflakeConnection]:
        if self.connector == "sqlalchemy":
            from snowflake.sqlalchemy import URL
            from sqlalchemy import create_engine, event

            engine = create_engine(
                URL(**self._sqlalchemy_connection_args),
                connect_args=self._sqlalchemy_engine_args,
            )

            # Attach a listener to the connection which will add a comment to any SQL query
            # executed through the connection. This comment will be used to identify the query
            # when later attributing cost.
            @event.listens_for(engine, "before_cursor_execute", retval=True)
            def comment_sql(conn, cursor, statement, parameters, context, executemany):
                context, _ = get_current_context_and_asset_key_or_warn()
                if not context:
                    return statement, parameters

                statement = meter_snowflake_query(
                    context, statement, associated_asset_key=asset_key
                )
                return statement, parameters

            conn = engine.raw_connection() if raw_conn else engine.connect()

            yield conn
            conn.close()
            engine.dispose()
        else:
            conn = WrappedSnowflakeConnection(asset_key=asset_key, **self._connection_args)

            yield conn
            if not self.autocommit:
                conn.commit()
            conn.close()


class InsightsSnowflakeConnection(SnowflakeConnection):
    def execute_query(
        self,
        sql: str,
        parameters: Sequence[Any] | Mapping[Any, Any] | None = None,
        fetch_results: bool = False,
        use_pandas_result: bool = False,
    ):
        context, asset_key = get_current_context_and_asset_key_or_warn()
        if not context:
            return super().execute_query(sql, parameters, fetch_results, use_pandas_result)

        return super().execute_query(
            meter_snowflake_query(context, sql, associated_asset_key=asset_key),
            parameters,
            fetch_results,
            use_pandas_result,
        )

    def execute_queries(
        self,
        sql_queries: Sequence[str],
        parameters: Sequence[Any] | Mapping[Any, Any] | None = None,
        fetch_results: bool = False,
        use_pandas_result: bool = False,
    ) -> Sequence[Any] | None:
        context, asset_key = get_current_context_and_asset_key_or_warn()

        if not context:
            return super().execute_queries(
                sql_queries, parameters, fetch_results, use_pandas_result
            )

        return super().execute_queries(
            [
                meter_snowflake_query(context, sql, associated_asset_key=asset_key)
                for sql in sql_queries
            ],
            parameters,
            fetch_results,
            use_pandas_result,
        )



import base64
from io import BytesIO

import matplotlib.pyplot as plt
import pandas as pd
import requests
from dagster import MetadataValue, OpExecutionContext, asset
from wordcloud import STOPWORDS, WordCloud


@asset(group_name="hackernews", compute_kind="HackerNews API")
def hackernews_topstory_ids() -> list[int]:
    """Get up to 500 top stories from the HackerNews topstories endpoint.

    API Docs: https://github.com/HackerNews/API#new-top-and-best-stories
    """
    newstories_url = "https://hacker-news.firebaseio.com/v0/topstories.json"
    top_500_newstories = requests.get(newstories_url).json()
    return top_500_newstories


@asset(group_name="hackernews", compute_kind="HackerNews API")
def hackernews_topstories(
    context: OpExecutionContext, hackernews_topstory_ids: list[int]
) -> pd.DataFrame:
    """Get items based on story ids from the HackerNews items endpoint. It may take 1-2 minutes to fetch all 500 items.

    API Docs: https://github.com/HackerNews/API#items
    """
    results = []
    for item_id in hackernews_topstory_ids:
        item = requests.get(f"https://hacker-news.firebaseio.com/v0/item/{item_id}.json").json()
        results.append(item)
        if len(results) % 20 == 0:
            context.log.info(f"Got {len(results)} items so far.")

    df = pd.DataFrame(results)

    # Dagster supports attaching arbitrary metadata to asset materializations. This metadata will be
    # shown in the run logs and also be displayed on the "Activity" tab of the "Asset Details" page in the UI.
    # This metadata would be useful for monitoring and maintaining the asset as you iterate.
    # Read more about in asset metadata in https://docs.dagster.io/concepts/assets/software-defined-assets#recording-materialization-metadata
    context.add_output_metadata(
        {
            "num_records": len(df),
            "preview": MetadataValue.md(df.head().to_markdown()),
        }
    )
    return df


@asset(group_name="hackernews", compute_kind="Plot")
def hackernews_topstories_word_cloud(
    context: OpExecutionContext, hackernews_topstories: pd.DataFrame
) -> bytes:
    """Exploratory analysis: Generate a word cloud from the current top 500 HackerNews top stories.
    Embed the plot into a Markdown metadata for quick view.

    Read more about how to create word clouds in http://amueller.github.io/word_cloud/.
    """
    stopwords = set(STOPWORDS)
    stopwords.update(["Ask", "Show", "HN"])
    titles_text = " ".join([str(item) for item in hackernews_topstories["title"]])
    titles_cloud = WordCloud(stopwords=stopwords, background_color="white").generate(titles_text)

    # Generate the word cloud image
    plt.figure(figsize=(8, 8), facecolor=None)
    plt.imshow(titles_cloud, interpolation="bilinear")
    plt.axis("off")
    plt.tight_layout(pad=0)

    # Save the image to a buffer and embed the image into Markdown content for quick view
    buffer = BytesIO()
    plt.savefig(buffer, format="png")
    image_data = base64.b64encode(buffer.getvalue())
    md_content = f"![img](data:image/png;base64,{image_data.decode()})"

    # Attach the Markdown content as metadata to the asset
    # Read about more metadata types in https://docs.dagster.io/_apidocs/ops#metadata-types
    context.add_output_metadata({"plot": MetadataValue.md(md_content)})

    return image_data



from dagster import (
    Definitions,
    ScheduleDefinition,
    define_asset_job,
    load_assets_from_package_module,
)

from quickstart_etl import assets  # pyright: ignore[reportMissingImports]

daily_refresh_schedule = ScheduleDefinition(
    job=define_asset_job(name="all_assets_job"), cron_schedule="0 0 * * *"
)

defs = Definitions(
    assets=load_assets_from_package_module(assets), schedules=[daily_refresh_schedule]
)



from dagster_cloud_cli.entrypoint import app
from typer import Typer
from typer.models import CommandInfo, TyperInfo


def _gather_groups_and_commands(t: Typer) -> tuple[list[TyperInfo], list[CommandInfo]]:
    groups = t.registered_groups
    commands = t.registered_commands

    for subgroup in t.registered_groups:
        if subgroup.typer_instance:
            sg_groups, sg_commands = _gather_groups_and_commands(subgroup.typer_instance)
            groups.extend(sg_groups)
            commands.extend(sg_commands)

    return groups, commands


def _group_has_help(info: TyperInfo):
    instance_has_help = info.typer_instance and _group_has_help(info.typer_instance.info)
    return info.help or info.short_help or info.hidden or instance_has_help


def _command_has_help(info: CommandInfo):
    return info.help or info.short_help or info.hidden or (info.callback and info.callback.__doc__)


def test_all_commands_and_groups_have_help_text():
    """Validates that all Typer commands and command groups have help text defined."""
    groups, commands = _gather_groups_and_commands(app)

    for group in groups:
        assert _group_has_help(group), f"Group {group.name} has no helptext"
    for command in commands:
        assert _command_has_help(command), f"Command {command.name} has no helptext"



import time

import dagster._check as check
import pytest
from dagster._core.events import (
    DagsterEvent,
    DagsterEventType,
    EngineEventData,
    HookErroredData,
    JobCanceledData,
    JobFailureData,
)
from dagster._core.events.log import EventLogEntry
from dagster._core.execution.plan.objects import StepFailureData, StepRetryData, UserFailureData
from dagster._core.test_utils import environ
from dagster._serdes import serialize_value
from dagster._utils.error import SerializableErrorInfo
from dagster_cloud.storage.event_logs.utils import (
    _get_maximum_event_message_characters,
    truncate_event,
)

BIG_STACK_TRACE = SerializableErrorInfo(
    message="Error occurred",
    stack=["a" * 100000],
    cls_name="Exception",
    cause=None,
)


def test_truncate_event():
    event = EventLogEntry(
        error_info=None,
        user_message="foobar",
        level="debug",
        run_id="foo",
        timestamp=time.time(),
        dagster_event=None,
    )
    assert truncate_event(event) == event

    event = EventLogEntry(
        error_info=None,
        level="debug",
        user_message="a" * 100000,
        run_id="foo",
        timestamp=time.time(),
        dagster_event=None,
    )
    truncated_event = truncate_event(event)
    assert len(truncated_event.user_message) < _get_maximum_event_message_characters() + 100
    assert truncated_event.user_message.startswith("[TRUNCATED")
    assert truncated_event.user_message.endswith("aaa [TRUNCATED]")


@pytest.fixture
def error_size():
    with environ({"DAGSTER_CLOUD_MAXIMUM_EVENT_ERROR_SIZE": "1000"}):
        yield 1000


@pytest.mark.parametrize(
    "event_type,event_specific_data,expected_truncations",
    [
        (
            DagsterEventType.STEP_FAILURE,
            StepFailureData(
                error=BIG_STACK_TRACE,
                user_failure_data=UserFailureData(label="K8sError"),
            ),
            ["stack"],
        ),
        (
            DagsterEventType.RUN_FAILURE,
            JobFailureData(
                error=BIG_STACK_TRACE,
                first_step_failure_event=DagsterEvent(
                    event_type_value="STEP_FAILURE",
                    job_name="my_job",
                    step_key="my_op",
                    message="Job failed",
                    event_specific_data=StepFailureData(
                        error=BIG_STACK_TRACE,
                        user_failure_data=UserFailureData(label="K8sError"),
                    ),
                ),
            ),
            ["stack", "stack"],
        ),
        (
            DagsterEventType.ENGINE_EVENT,
            EngineEventData(
                error=BIG_STACK_TRACE,
                marker_start=None,
                marker_end=None,
            ),
            ["stack"],
        ),
        (
            DagsterEventType.RUN_CANCELED,
            JobCanceledData(
                error=BIG_STACK_TRACE,
            ),
            ["stack"],
        ),
        (
            DagsterEventType.HOOK_ERRORED,
            HookErroredData(
                error=BIG_STACK_TRACE,
            ),
            ["stack"],
        ),
        (
            DagsterEventType.STEP_UP_FOR_RETRY,
            StepRetryData(
                error=BIG_STACK_TRACE,
                seconds_to_wait=10,
            ),
            ["stack"],
        ),
    ],
)
def test_truncated_event_errors(error_size, event_type, event_specific_data, expected_truncations):
    event = EventLogEntry(
        error_info=None,
        level="error",
        user_message="Job failed",
        run_id="foo",
        timestamp=time.time(),
        dagster_event=DagsterEvent(
            event_type_value=event_type.value,
            job_name="my_job",
            step_key="my_op",
            message="Job failed",
            event_specific_data=event_specific_data,
        ),
    )

    error = check.not_none(check.not_none(event.dagster_event).event_specific_data.error)  # type:ignore
    assert len(serialize_value(error)) > error_size

    truncations = []
    truncated_event = truncate_event(event, truncations=truncations)
    assert truncations == expected_truncations

    truncated_error = check.not_none(
        check.not_none(truncated_event.dagster_event).event_specific_data.error  # type:ignore
    )
    assert len(serialize_value(truncated_error)) < error_size + 100

    if event_type == DagsterEventType.RUN_FAILURE:
        error = event.dagster_event.event_specific_data.first_step_failure_event.event_specific_data.error  # type:ignore
        assert len(serialize_value(error)) > error_size

        truncated_event = truncate_event(event)

        truncated_error = truncated_event.dagster_event.event_specific_data.first_step_failure_event.event_specific_data.error  # type:ignore

        assert len(serialize_value(truncated_error)) < error_size + 100



import pytest
from dagster._utils.error import SerializableErrorInfo
from dagster_cloud.util.container_resources import (
    interpret_ecs_cpu_str_as_millicpus,
    interpret_ecs_mem_str_as_bytes,
    interpret_k8s_cpu_str_as_millicpus,
    interpret_k8s_mem_str_as_bytes,
)
from dagster_cloud.util.errors import remove_dagster_framework_lines_from_serializable_exc_info


def test_resource_limit_interpreters():
    for fn in [
        interpret_ecs_cpu_str_as_millicpus,
        interpret_ecs_mem_str_as_bytes,
        interpret_k8s_cpu_str_as_millicpus,
        interpret_k8s_mem_str_as_bytes,
    ]:
        assert fn(None) is None

    assert interpret_k8s_mem_str_as_bytes("1") == 1
    assert interpret_k8s_mem_str_as_bytes("1Mi") == 1024**2
    assert interpret_k8s_mem_str_as_bytes("1M") == 1000**2
    assert interpret_k8s_mem_str_as_bytes("1Gi") == 1024**3
    assert interpret_k8s_mem_str_as_bytes("1G") == 1000**3
    assert interpret_k8s_mem_str_as_bytes("1Ti") == 1024**4
    assert interpret_k8s_mem_str_as_bytes("1T") == 1000**4
    assert interpret_k8s_mem_str_as_bytes("1Pi") == 1024**5
    assert interpret_k8s_mem_str_as_bytes("1P") == 1000**5
    assert interpret_k8s_mem_str_as_bytes("1Ei") == 1024**6
    assert interpret_k8s_mem_str_as_bytes("1E") == 1000**6
    assert interpret_k8s_mem_str_as_bytes("1K") == 1000
    assert interpret_k8s_mem_str_as_bytes("1Ki") == 1024

    with pytest.raises(Exception):
        interpret_k8s_mem_str_as_bytes("1KB")

    assert interpret_k8s_cpu_str_as_millicpus("10") == 10000
    assert interpret_k8s_cpu_str_as_millicpus("1m") == 1
    assert interpret_k8s_cpu_str_as_millicpus("0.1") == 100

    with pytest.raises(Exception):
        interpret_k8s_cpu_str_as_millicpus("1K")

    assert interpret_ecs_mem_str_as_bytes("1") == 1024**2
    assert interpret_ecs_mem_str_as_bytes("1GB") == 1024**3

    with pytest.raises(Exception):
        interpret_ecs_mem_str_as_bytes("1Gi")

    assert interpret_ecs_cpu_str_as_millicpus("10") == 10
    assert interpret_ecs_cpu_str_as_millicpus("1") == 1

    with pytest.raises(Exception):
        interpret_ecs_cpu_str_as_millicpus("1m")

    with pytest.raises(Exception):
        interpret_ecs_cpu_str_as_millicpus("0.1")

    for test_str in [
        "0.25vcpu",
        "0.25VCPU",
        "0.25vCPU",
        "0.25VCpu",
        "0.25vCpu",
        "0.25 vcpu",
        "0.25 VCPU",
        "0.25 vCPU",
        "0.25 VCpu",
        "0.25 vCpu",
    ]:
        assert interpret_ecs_cpu_str_as_millicpus(test_str) == 250


from dagster import deserialize_value, op

SERIALIZED_PROD_ERROR_WITH_FRAMEWORK_LINES = """{"__class__": "SerializableErrorInfo", "cause": null, "cls_name": "Exception", "context": {"__class__": "SerializableErrorInfo", "cause": null, "cls_name": "Exception", "context": null, "message": "Exception: Sling command failed:\\n[]\\n", "stack": ["  File \\"/usr/local/lib/python3.11/site-packages/sling/__init__.py\\", line 452, in _run\\n    for line in _exec_cmd(cmd, env=env, stdin=stdin):\\n", "  File \\"/usr/local/lib/python3.11/site-packages/sling/__init__.py\\", line 520, in _exec_cmd\\n    raise Exception(f'Sling command failed:\\\\n{lines}')\\n"]}, "message": "Exception: \\n\\u001b[90m7:40PM\\u001b[0m \\u001b[32mINF\\u001b[0m Sling Replication [1 streams] | CLOUD_PRODUCTION_MAIN -> SLING_DB_MAIN\\n\\u001b[90m7:40PM\\u001b[0m \\u001b[32mINF\\u001b[0m [1 / 1] running stream public.event_logs_view\\n\\u001b[90m7:40PM\\u001b[0m \\u001b[32mINF\\u001b[0m connecting to source database (postgres)\\n\\u001b[90m7:40PM\\u001b[0m \\u001b[32mINF\\u001b[0m \\u001b[31mexecution failed\\u001b[0m\\n\\n\\u001b[31mfatal:\\n~ failure running replication (see docs @ https://docs.slingdata.io/sling-cli)\\n--------------------------- public.event_logs_view ---------------------------\\n~ could not connect to database\\npq: canceling statement due to conflict with recovery\\u001b[0m\\n\\u001b[90m7:40PM\\u001b[0m \\u001b[32mINF\\u001b[0m \\u001b[31m~ could not connect to database\\npq: canceling statement due to conflict with recovery\\u001b[0m\\n\\n\\u001b[35mPerhaps adjusting the `max_standby_archive_delay` and `max_standby_streaming_delay` settings in the source PG Database could help. See https://stackoverflow.com/questions/14592436/postgresql-error-canceling-statement-due-to-conflict-with-recovery\\u001b[0m\\n\\n\\u001b[90m7:40PM\\u001b[0m \\u001b[32mINF\\u001b[0m Sling Replication Completed in 0s | CLOUD_PRODUCTION_MAIN -> SLING_DB_MAIN | \\u001b[32m0 Successes\\u001b[0m | \\u001b[31m1 Failures\\u001b[0m\\n\\nSling command failed:\\n[]\\n", "stack": ["  File \\"/usr/local/lib/python3.11/site-packages/dagster/_core/execution/plan/utils.py\\", line 54, in op_execution_error_boundary\\n    yield\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster/_utils/__init__.py\\", line 482, in iterate_with_context\\n    next_output = next(iterator)\\n                  ^^^^^^^^^^^^^^\\n", "  File \\"/opt/python_modules/dagster-open-platform/dagster_open_platform/sling/assets.py\\", line 69, in cloud_product_main_event_log\\n    yield from embedded_elt.replicate(context=context).fetch_column_metadata().fetch_row_count()\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster_embedded_elt/sling/sling_event_iterator.py\\", line 189, in __next__\\n    return next(self._inner_iterator)\\n           ^^^^^^^^^^^^^^^^^^^^^^^^^^\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster_embedded_elt/sling/sling_event_iterator.py\\", line 229, in _fetch_row_count\\n    for event in self:\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster_embedded_elt/sling/sling_event_iterator.py\\", line 189, in __next__\\n    return next(self._inner_iterator)\\n           ^^^^^^^^^^^^^^^^^^^^^^^^^^\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster_embedded_elt/sling/sling_event_iterator.py\\", line 206, in _fetch_column_metadata\\n    for event in self:\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster_embedded_elt/sling/sling_event_iterator.py\\", line 189, in __next__\\n    return next(self._inner_iterator)\\n           ^^^^^^^^^^^^^^^^^^^^^^^^^^\\n", "  File \\"/usr/local/lib/python3.11/site-packages/dagster_embedded_elt/sling/resources.py\\", line 388, in _replicate\\n    results = sling._run(  # noqa\\n              ^^^^^^^^^^^^^^^^^^^\\n", "  File \\"/usr/local/lib/python3.11/site-packages/sling/__init__.py\\", line 465, in _run\\n    raise Exception('\\\\n'.join(lines))\\n"]}"""

LOCAL_DEV_ERROR_WITH_FRAMEWORK_LINES = """{"__class__": "SerializableErrorInfo", "cause": null, "cls_name": "Exception", "context": {"__class__": "SerializableErrorInfo", "cause": null, "cls_name": "Exception", "context": null, "message": "Exception: INNER EXCEPTION\\n", "stack": ["  File \\"hi.py\\", line 21, in my_asset\\n    raise Exception(\\"INNER EXCEPTION\\")\\n"]}, "message": "Exception: OUTER EXCEPTION\\n", "stack": ["  File \\"/Users/dgibson/dagster/python_modules/dagster/dagster/_core/execution/plan/utils.py\\", line 54, in op_execution_error_boundary\\n    yield\\n", "  File \\"/Users/dgibson/dagster/python_modules/dagster/dagster/_utils/__init__.py\\", line 490, in iterate_with_context\\n    next_output = next(iterator)\\n", "  File \\"/Users/dgibson/dagster/python_modules/dagster/dagster/_core/execution/plan/compute_generator.py\\", line 140, in _coerce_op_compute_fn_to_iterator\\n    result = invoke_compute_fn(\\n", "  File \\"/Users/dgibson/dagster/python_modules/dagster/dagster/_core/execution/plan/compute_generator.py\\", line 128, in invoke_compute_fn\\n    return fn(context, **args_to_pass) if context_arg_provided else fn(**args_to_pass)\\n", "  File \\"hi.py\\", line 23, in my_asset\\n    raise Exception(\\"OUTER EXCEPTION\\")\\n"]}"""


def _helper_fn():
    raise Exception("Inner Exception")


@op
def my_op():
    try:
        _helper_fn()
    except:
        raise Exception("Outer exception")


def test_remove_dagster_framework_lines_from_prod_stack_trace(snapshot):
    # Snapshotting a constant looks a little pointless, but shows the before and after
    # more clearly in the snapshot file
    snapshot.assert_match(
        deserialize_value(SERIALIZED_PROD_ERROR_WITH_FRAMEWORK_LINES, SerializableErrorInfo)
    )

    serializable_error = deserialize_value(
        SERIALIZED_PROD_ERROR_WITH_FRAMEWORK_LINES, SerializableErrorInfo
    )

    removed_dagster_framework_error = remove_dagster_framework_lines_from_serializable_exc_info(
        serializable_error
    )

    snapshot.assert_match(removed_dagster_framework_error)


def test_remove_dagster_framework_lines_from_local_dev_stack_trace(snapshot):
    snapshot.assert_match(
        deserialize_value(LOCAL_DEV_ERROR_WITH_FRAMEWORK_LINES, SerializableErrorInfo)
    )

    serializable_error = deserialize_value(
        LOCAL_DEV_ERROR_WITH_FRAMEWORK_LINES, SerializableErrorInfo
    )

    removed_dagster_framework_error = remove_dagster_framework_lines_from_serializable_exc_info(
        serializable_error
    )

    snapshot.assert_match(removed_dagster_framework_error)



import importlib
import os
import shutil
import sys
from pathlib import Path
from typing import cast
from unittest import mock

import pytest
import requests
from dagster import AssetsDefinition, materialize
from dagster._core.test_utils import environ
from dagster_cloud_cli.core.artifacts import (
    download_deployment_artifact,
    download_organization_artifact,
    upload_deployment_artifact,
    upload_organization_artifact,
)
from dagster_cloud_cli.entrypoint import app
from dagster_dbt import DbtCliResource, DbtProject
from dagster_dbt.cli.app import app as dagster_dbt_app
from typer.testing import CliRunner

runner = CliRunner()


@pytest.fixture
def ci_env(agent_token):
    with environ(
        {
            "DAGSTER_CLOUD_ORGANIZATION": "acme",
            "DAGSTER_CLOUD_DEPLOYMENT": "prod",
            "DAGSTER_CLOUD_API_TOKEN": agent_token,
        }
    ):
        yield


@pytest.fixture
def mocked_dagit(dagit_test_client):
    with mock.patch(
        "dagster_cloud_cli.core.artifacts._dagster_cloud_http_client",
        return_value=dagit_test_client,
    ):
        yield


@pytest.fixture
def dbt_project_dir(tmp_path: Path) -> Path:
    dbt_project_dir = tmp_path.joinpath("test_jaffle_shop")
    shutil.copytree(src=Path(__file__).joinpath("..", "dbt_project").resolve(), dst=dbt_project_dir)

    return dbt_project_dir


def test_org_scoped(mocked_dagit, ci_env, tmp_path: Path):
    content = "yikes"
    origin_file = tmp_path.joinpath("origin.txt")
    origin_file.write_text(content)

    key = "o-artifact"

    upload_organization_artifact(key=key, path=origin_file)

    copy_path = tmp_path.joinpath("copy.txt")

    download_organization_artifact(key=key, path=copy_path)

    assert copy_path.read_text() == content
    with pytest.raises(requests.exceptions.HTTPError, match="404"):
        download_deployment_artifact(key=key, path=copy_path)


def test_deploy_scoped(mocked_dagit, ci_env, tmp_path: Path):
    content = "yikes"
    origin_file = tmp_path.joinpath("origin.txt")
    origin_file.write_text(content)

    key = "d-artifact"

    upload_deployment_artifact(key=key, path=origin_file)

    copy_path = tmp_path.joinpath("copy.txt")

    download_deployment_artifact(key=key, path=copy_path)

    assert copy_path.read_text() == content

    with pytest.raises(requests.exceptions.HTTPError, match="404"):
        download_organization_artifact(key=key, path=copy_path)


def test_missing_file(ci_env, tmp_path: Path):
    origin_file = tmp_path.joinpath("missing.txt")
    with pytest.raises(FileNotFoundError):
        upload_organization_artifact(key="the-artifact", path=origin_file)


def test_org_mismatch(mocked_dagit, tmp_path: Path, agent_token):
    with environ(
        {
            "DAGSTER_CLOUD_ORGANIZATION": "hooli",
            "DAGSTER_CLOUD_API_TOKEN": agent_token,  # acme token
        }
    ):
        content = "yikes"
        origin_file = tmp_path.joinpath("origin.txt")
        origin_file.write_text(content)

        key = "o-artifact"
        with pytest.raises(Exception, match="401"):
            upload_organization_artifact(key=key, path=origin_file)


def test_manage_state(
    monkeypatch: pytest.MonkeyPatch,
    mocked_dagit,
    agent_token: str,
    dbt_project_dir: Path,
    tmp_path: Path,
) -> None:
    project_name = "jaffle_dagster"
    dagster_project_dir = dbt_project_dir.joinpath(project_name)
    dbt_project_file_path = dagster_project_dir.joinpath(project_name, "project.py")

    monkeypatch.chdir(dbt_project_dir)
    sys.path.append(os.fspath(dbt_project_dir))

    monkeypatch.setenv("DAGSTER_CLOUD_ORGANIZATION", "acme")
    monkeypatch.setenv("DAGSTER_CLOUD_API_TOKEN", agent_token)

    # Scaffold the project
    result = runner.invoke(
        dagster_dbt_app,
        [
            "project",
            "scaffold",
            "--project-name",
            project_name,
            "--dbt-project-dir",
            os.fspath(dbt_project_dir),
            "--use-experimental-dbt-state",
        ],
    )
    assert result.exit_code == 0

    # Build the manifest
    result = runner.invoke(
        dagster_dbt_app,
        ["project", "prepare-and-package", "--file", os.fspath(dbt_project_file_path)],
    )

    assert result.exit_code == 0, result.output

    dagster_cloud_state_dir = tmp_path.joinpath("build")
    dagster_cloud_state_dir.mkdir()

    # prod ci run
    monkeypatch.setenv("DAGSTER_CLOUD_DEPLOYMENT", "prod")

    result = runner.invoke(
        app,
        [
            "ci",
            "init",
            "--statedir",
            os.fspath(dagster_cloud_state_dir),
            "--project-dir",
            os.fspath(dbt_project_dir),
        ],
    )
    assert result.exit_code == 0, result.output

    result = runner.invoke(
        app,
        [
            "ci",
            "dagster-dbt",
            "project",
            "manage-state",
            "--file",
            os.fspath(dbt_project_file_path),
            "--statedir",
            os.fspath(dagster_cloud_state_dir),
        ],
    )
    assert result.exit_code == 0, result.output
    assert "Upload complete" in result.output

    scaffold_defs_module = importlib.import_module(f"{project_name}.{project_name}.definitions")
    my_dbt_assets: AssetsDefinition = getattr(scaffold_defs_module, "jaffle_shop_dbt_assets")
    project = cast("DbtProject", getattr(scaffold_defs_module, "jaffle_shop_project"))
    dbt = DbtCliResource(project_dir=project)

    assert dbt.state_path

    # prod full materialization
    result = materialize([my_dbt_assets], resources={"dbt": dbt})
    assert result.success

    # branch ci run
    monkeypatch.setenv("DAGSTER_CLOUD_DEPLOYMENT", "staging")
    monkeypatch.setenv("DAGSTER_DBT_JAFFLE_SCHEMA", "staging")

    # Running in staging should fail because the state directory is empty, so there is no --defer.
    result = materialize(
        [my_dbt_assets], selection="orders", resources={"dbt": dbt}, raise_on_error=False
    )
    assert not result.success

    with mock.patch(
        "dagster_cloud_cli.commands.ci.create_or_update_deployment_from_context",
        return_value="branch-dep-name",
    ):
        result = runner.invoke(
            app,
            [
                "ci",
                "init",
                "--statedir",
                os.fspath(dagster_cloud_state_dir),
                "--project-dir",
                os.fspath(dbt_project_dir),
            ],
        )
        assert result.exit_code == 0, result.output

        result = runner.invoke(
            app,
            [
                "ci",
                "dagster-dbt",
                "project",
                "manage-state",
                "--file",
                os.fspath(dbt_project_file_path),
                "--statedir",
                os.fspath(dagster_cloud_state_dir),
            ],
        )
        assert result.exit_code == 0, result.output
        assert "Download complete" in result.output

    # Once the state directory is populated, the subselected asset can be produced.
    result = materialize([my_dbt_assets], selection="orders", resources={"dbt": dbt})
    assert result.success



"""(very) small perf testing script for batcher that spins up
a large number of threads and tries to detect overhead of lock
contention in the batcher. the handler sleeps for 100ms
imitating a db call, and so you should be able to compute what
the expected runtime should be (for example, with n=10k,
max_workers=100, max_batch_size=100, it should take 100 * 100ms = 10s
(the batcher coalesces work across threads)).
"""

import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager

import structlog
from dagster_cloud.batching.batcher import Batcher
from typer import Typer

logger = structlog.get_logger(__name__)

app = Typer()


def load_test(n: int, max_batch_size: int):
    def handler(xs: list[int]) -> list[int]:
        time.sleep(0.1)
        return [x + 1 for x in xs]

    batcher: Batcher[int, int] = Batcher(
        "test",
        handler,
        max_batch_size=max_batch_size,
    )

    with ThreadPoolExecutor(max_workers=100) as pool:
        result = list(pool.map(batcher.submit, range(n)))

    assert result == list(range(1, n + 1))


@contextmanager
def timeit(n: int):
    s = time.monotonic()
    yield
    e = time.monotonic()
    logger.info(f"took {e - s} seconds for {n} iterations")


@app.command()
def command(n: int, max_batch_size: int = 100):
    with timeit(n):
        load_test(n, max_batch_size)


if __name__ == "__main__":
    app()



import logging
import re
import time
from contextlib import contextmanager
from unittest import mock
from unittest.mock import Mock

import pytest
from dagster._core.test_utils import instance_for_test
from dagster._utils.merger import merge_dicts
from dagster_cloud.workspace.kubernetes.launcher import (
    DEFAULT_DEPLOYMENT_STARTUP_TIMEOUT,
    DEFAULT_IMAGE_PULL_GRACE_PERIOD,
    K8sUserCodeLauncher,
)
from dagster_cloud.workspace.kubernetes.utils import (
    construct_code_location_deployment,
    construct_code_location_service,
    get_deployment_failure_debug_info,
    get_k8s_human_readable_label,
    unique_k8s_resource_name,
)
from dagster_cloud.workspace.user_code_launcher import (
    DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT,
    UserCodeLauncherEntry,
)
from dagster_cloud_cli.core.workspace import CodeLocationDeployData
from dagster_k8s.container_context import K8sContainerContext
from dagster_k8s.job import UserDefinedDagsterK8sConfig


def test_config():
    assert K8sUserCodeLauncher.config_type()


# From the failure you get when you make a k8s deployment with an invalid nmae
K8S_SERVICE_REGEX = "[a-z]([-a-z0-9]*[a-z0-9])?"

MINIMAL_KUBECONFIG_CONTENT = """
apiVersion: v1
kind: Config

current-context: fake-context
contexts:
  - context:
      cluster: fake-cluster
    name: fake-context
clusters:
  - cluster: {}
    name: fake-cluster
"""


@pytest.fixture
def kubeconfig_file(tmp_path):
    """Returns a str file path for a minimal kubeconfig file in the default location (~/.kube/config)."""
    dir_path = tmp_path / ".kube"
    dir_path.mkdir()
    config_path = dir_path / "config"
    config_path.write_text(MINIMAL_KUBECONFIG_CONTENT)
    return str(config_path)


@contextmanager
def k8s_instance(user_code_launcher_overrides=None):
    with mock.patch("kubernetes.config.load_incluster_config"):
        with instance_for_test(
            {
                "instance_class": {
                    "module": "dagster_cloud",
                    "class": "DagsterCloudAgentInstance",
                },
                "user_code_launcher": {
                    "module": "dagster_cloud.workspace.kubernetes",
                    "class": "K8sUserCodeLauncher",
                    "config": merge_dicts(
                        {
                            "dagster_home": "MY_DAGSTER_HOME",
                            "instance_config_map": "MY_INSTANCE_CONFIG_MAP",
                            "service_account_name": "MY_SERVICE_ACCOUNT_NAME",
                        },
                        user_code_launcher_overrides or {},
                    ),
                },
                "dagster_cloud_api": {
                    "url": "http://localhost:2874",
                    "agent_token": "FAKE_TOKEN",
                },
                "compute_logs": {
                    "module": "dagster.core.storage.noop_compute_log_manager",
                    "class": "NoOpComputeLogManager",
                },
            }
        ) as instance:
            yield instance


def test_default_instance():
    with k8s_instance() as instance:
        assert (
            instance.user_code_launcher._deployment_startup_timeout  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_DEPLOYMENT_STARTUP_TIMEOUT
        )
        assert (
            instance.user_code_launcher._server_process_startup_timeout  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT
        )

        assert (
            instance.user_code_launcher._image_pull_grace_period  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_IMAGE_PULL_GRACE_PERIOD
        )

        assert instance.user_code_launcher._labels == {}  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher._resources == {}  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_timeout_overrides():
    with k8s_instance(
        {
            "deployment_startup_timeout": 123,
            "server_process_startup_timeout": 456,
            "image_pull_grace_period": 789,
        }
    ) as instance:
        assert instance.user_code_launcher._deployment_startup_timeout == 123  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher._server_process_startup_timeout == 456  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher._image_pull_grace_period == 789  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_labels():
    with k8s_instance({"labels": {"foo": "bar"}}) as instance:
        assert instance.user_code_launcher._labels == {"foo": "bar"}  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_env_vars():
    env_vars = ["FOO_ENV_VAR", "BAR_ENV_VAR=BAR_VALUE"]

    with k8s_instance({"env_vars": env_vars}) as instance:
        assert instance.user_code_launcher._env_vars == env_vars  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_volumes_instance():
    with k8s_instance(
        {
            "volume_mounts": [{"name": "foo", "mountPath": "biz/buz", "subPath": "file.txt"}],
            "volumes": [
                {"name": "foo", "configMap": {"name": "settings-cm"}},
            ],
        }
    ) as instance:
        assert instance.user_code_launcher._volume_mounts == [  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            {
                "name": "foo",
                "mount_path": "biz/buz",
                "sub_path": "file.txt",
            }
        ]
        assert instance.user_code_launcher._volumes == [  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            {"name": "foo", "config_map": {"name": "settings-cm"}}
        ]


def test_resources_instance():
    resources = {
        "requests": {"cpu": "250m", "memory": "64Mi"},
        "limits": {"cpu": "500m", "memory": "2560Mi"},
    }
    with k8s_instance({"resources": resources}) as instance:
        assert instance.user_code_launcher._resources == resources  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]


def test_security_context_instance():
    sacred_rites_of_debugging = {"capabilities": {"add": ["SYS_PTRACE"]}}

    with k8s_instance({"security_context": sacred_rites_of_debugging}) as instance:
        assert (
            instance.user_code_launcher._security_context  # noqa: SLF001  # pyright: ignore[reportAttributeAccessIssue]
            == sacred_rites_of_debugging
        )


def test_sanitize_k8s_name():
    assert "foobar-sandbox-" in unique_k8s_resource_name("SandBox", "fOo_bAr")
    assert "!" not in unique_k8s_resource_name("sandbox", "sillyname!")
    assert len(unique_k8s_resource_name("sandbox", "extralong name" * 100)) <= 63
    assert "-location" not in unique_k8s_resource_name("hyphen", "-location")

    assert re.match(K8S_SERVICE_REGEX, unique_k8s_resource_name("sandbox", "foo_bar"))
    assert re.match(K8S_SERVICE_REGEX, unique_k8s_resource_name("22fast2furious", "33-iscompany"))

    assert "iscompany-22fast2furious" in unique_k8s_resource_name("22fast2furious", "33-iscompany")

    assert re.match(K8S_SERVICE_REGEX, unique_k8s_resource_name("222", "333"))


def test_get_k8s_human_readable_label():
    assert get_k8s_human_readable_label("foo_bar") == "foo_bar"
    assert "!" not in get_k8s_human_readable_label("sillyname!")


def test_construct_code_location_service():
    resource_name = unique_k8s_resource_name("sandbox", "biz.buz")

    container_context = K8sContainerContext(
        labels={"foo_label": "bar"},
        server_k8s_config=UserDefinedDagsterK8sConfig.from_dict(
            {
                "service_metadata": {
                    "annotations": {"foo_service": "bar"},
                    "labels": {"extra_label": "extra_value"},
                },
                "service_spec_config": {"cluster_ip": "None"},
            }
        ),
    )

    with k8s_instance() as instance:
        server_timestamp = time.time()

        obj = construct_code_location_service(
            "sandbox",
            "biz.buz",
            resource_name,
            container_context,
            instance,
            server_timestamp,
        ).to_dict()

        assert obj
        assert obj["metadata"]["name"] == resource_name
        assert obj["metadata"]["labels"]["foo_label"] == "bar"
        assert obj["metadata"]["labels"]["managed_by"] == "K8sUserCodeLauncher"
        assert obj["metadata"]["labels"]["location_name"] == "biz.buz"
        assert obj["metadata"]["labels"]["agent_id"] == instance.instance_uuid  # pyright: ignore[reportAttributeAccessIssue]
        assert obj["metadata"]["labels"]["server_timestamp"] == str(server_timestamp)
        assert obj["metadata"]["labels"]["extra_label"] == "extra_value"

        assert obj["metadata"]["annotations"] == {"foo_service": "bar"}
        assert obj["spec"]["cluster_ip"] == "None"


def test_construct_code_location_service_with_service_spec_config():
    resource_name = unique_k8s_resource_name("sandbox", "biz.buz")

    container_context = K8sContainerContext(
        server_k8s_config=UserDefinedDagsterK8sConfig.from_dict(
            {
                "service_spec_config": {"cluster_ip": "None"},
            }
        ),
    )

    with k8s_instance() as instance:
        server_timestamp = time.time()

        obj = construct_code_location_service(
            "sandbox",
            "biz.buz",
            resource_name,
            container_context,
            instance,
            server_timestamp,
        ).to_dict()

        assert obj
        assert obj["spec"]["cluster_ip"] == "None"
        assert obj["spec"]["selector"] == {"user-deployment": resource_name}
        assert obj["spec"]["ports"][0]["name"] == "grpc"
        assert obj["spec"]["ports"][0]["protocol"] == "TCP"


def test_construct_code_location_deployment():
    resource_name = unique_k8s_resource_name("sandbox", "foobar")

    container_context = K8sContainerContext(
        image_pull_policy="IfNotPresent",
        env_secrets=["dagster-cloud-agent-token"],
        env_config_maps=["user-config-map"],
        service_account_name="dagster",
        image_pull_secrets=[{"name": "test-image-pull-secret"}],
        volume_mounts=[{"name": "foo", "mount_path": "biz/buz", "sub_path": "file.txt"}],
        volumes=[
            {"name": "foo", "config_map": {"name": "settings-cm"}},
        ],
        labels={"foo_label": "bar"},
        resources={
            "requests": {"cpu": "250m", "memory": "64Mi"},
            "limits": {"cpu": "500m", "memory": "2560Mi"},
        },
    )

    with k8s_instance() as instance:
        server_timestamp = time.time()
        obj = construct_code_location_deployment(
            instance,
            deployment_name="sandbox",
            location_name="foobar_",
            k8s_deployment_name=resource_name,
            metadata=CodeLocationDeployData("bizbuz", package_name="blim"),
            container_context=container_context,
            args=["ls"],
            server_timestamp=server_timestamp,
        ).to_dict()

        assert obj
        assert obj["metadata"]["name"] == resource_name
        assert obj["metadata"]["labels"]["foo_label"] == "bar"
        assert obj["metadata"]["labels"]["managed_by"] == "K8sUserCodeLauncher"
        assert obj["metadata"]["labels"]["location_name"] == "foobar"
        assert obj["metadata"]["labels"]["agent_id"] == instance.instance_uuid  # pyright: ignore[reportAttributeAccessIssue]
        assert obj["metadata"]["labels"]["server_timestamp"] == str(server_timestamp)

        assert len(obj["spec"]["template"]["spec"]["containers"]) == 1

        assert obj["spec"]["template"]["metadata"]["labels"]["foo_label"] == "bar"
        assert obj["spec"]["template"]["metadata"]["labels"]["managed_by"] == "K8sUserCodeLauncher"
        assert obj["spec"]["template"]["metadata"]["labels"]["location_name"] == "foobar"
        assert obj["spec"]["template"]["metadata"]["labels"]["agent_id"] == instance.instance_uuid  # pyright: ignore[reportAttributeAccessIssue]

        assert obj["spec"]["template"]["spec"]["containers"][0]["image"] == "bizbuz"
        assert (
            obj["spec"]["template"]["spec"]["containers"][0]["image_pull_policy"] == "IfNotPresent"
        )

        assert len(obj["spec"]["template"]["spec"]["containers"][0]["env_from"]) == 2
        assert (
            obj["spec"]["template"]["spec"]["containers"][0]["env_from"][0]["config_map_ref"][
                "name"
            ]
            == "user-config-map"
        )
        assert (
            obj["spec"]["template"]["spec"]["containers"][0]["env_from"][1]["secret_ref"]["name"]
            == "dagster-cloud-agent-token"
        )

        assert obj["spec"]["template"]["spec"]["containers"][0]["args"] == ["ls"]
        assert obj["spec"]["template"]["spec"]["service_account_name"] == "dagster"
        assert obj["spec"]["template"]["spec"]["image_pull_secrets"] == [
            {"name": "test-image-pull-secret"}
        ]

        assert len(obj["spec"]["template"]["spec"]["volumes"]) == 1
        foo_volumes = [
            volume
            for volume in obj["spec"]["template"]["spec"]["volumes"]
            if volume["name"] == "foo"
        ]
        assert len(foo_volumes) == 1
        assert foo_volumes[0]["config_map"]["name"] == "settings-cm"

        assert len(obj["spec"]["template"]["spec"]["containers"][0]["volume_mounts"]) == 1
        foo_volumes_mounts = [
            volume
            for volume in obj["spec"]["template"]["spec"]["containers"][0]["volume_mounts"]
            if volume["name"] == "foo"
        ]
        assert len(foo_volumes_mounts) == 1

        assert obj["spec"]["template"]["spec"]["scheduler_name"] is None

        with pytest.raises(
            Exception, match=r"Unexpected keys in model class V1Volume: {'invalid_key'}"
        ):
            container_context = K8sContainerContext(
                image_pull_policy="IfNotPresent",
                env_secrets=["dagster-cloud-agent-token"],
                env_config_maps=["user-config-map"],
                service_account_name="dagster",
                image_pull_secrets=[{"name": "test-image-pull-secret"}],
                volume_mounts=[{"name": "foo", "mount_path": "biz/buz", "sub_path": "file.txt"}],
                volumes=[
                    {"name": "foo", "invalid_key": "settings-secret"},
                ],
                labels={},
                resources={},
            )
            construct_code_location_deployment(  # pyright: ignore[reportCallIssue]
                instance,
                deployment_name="sandbox",
                location_name="foobar_",
                k8s_deployment_name=resource_name,
                metadata=CodeLocationDeployData("bizbuz", package_name="blim"),
                container_context=container_context,
                server_timestamp=time.time(),
            )


def test_construct_code_location_deployment_scheduler_name():
    resource_name = unique_k8s_resource_name("sandbox", "foobar")
    with k8s_instance() as instance:
        container_context = K8sContainerContext(
            image_pull_policy="IfNotPresent",
            env_secrets=["dagster-cloud-agent-token"],
            env_config_maps=["user-config-map"],
            service_account_name="dagster",
            image_pull_secrets=[{"name": "test-image-pull-secret"}],
            volume_mounts=[{"name": "foo", "mount_path": "biz/buz", "sub_path": "file.txt"}],
            volumes=[
                {"name": "foo", "config_map": {"name": "settings-cm"}},
            ],
            labels={"foo_label": "bar"},
            resources={
                "requests": {"cpu": "250m", "memory": "64Mi"},
                "limits": {"cpu": "500m", "memory": "2560Mi"},
            },
            scheduler_name="test-scheduler",
        )
        obj = construct_code_location_deployment(
            instance,
            deployment_name="sandbox",
            location_name="foobar_",
            k8s_deployment_name=resource_name,
            metadata=CodeLocationDeployData("bizbuz", package_name="blim"),
            container_context=container_context,
            args=["ls"],
            server_timestamp=time.time(),
        ).to_dict()

        assert obj
        assert obj["spec"]["template"]["spec"]["scheduler_name"] == "test-scheduler"


def test_construct_code_location_deployment_security_context():
    resource_name = unique_k8s_resource_name("sandbox", "foobar")
    with k8s_instance() as instance:
        container_context = K8sContainerContext(
            image_pull_policy="IfNotPresent",
            env_secrets=["dagster-cloud-agent-token"],
            env_config_maps=["user-config-map"],
            service_account_name="dagster",
            image_pull_secrets=[{"name": "test-image-pull-secret"}],
            volume_mounts=[{"name": "foo", "mount_path": "biz/buz", "sub_path": "file.txt"}],
            volumes=[
                {"name": "foo", "config_map": {"name": "settings-cm"}},
            ],
            labels={"foo_label": "bar"},
            resources={
                "requests": {"cpu": "250m", "memory": "64Mi"},
                "limits": {"cpu": "500m", "memory": "2560Mi"},
            },
            security_context={"capabilities": {"add": ["SYS_PTRACE"]}},
        )
        obj = construct_code_location_deployment(
            instance,
            deployment_name="sandbox",
            location_name="foobar_",
            k8s_deployment_name=resource_name,
            metadata=CodeLocationDeployData("bizbuz", package_name="blim"),
            container_context=container_context,
            args=["ls"],
            server_timestamp=time.time(),
        ).to_dict()

        assert obj
        assert obj["spec"]["template"]["spec"]["containers"][0]["security_context"]["capabilities"][
            "add"
        ] == ["SYS_PTRACE"]


def test_construct_code_location_deployment_with_raw_k8s_config():
    resource_name = unique_k8s_resource_name("sandbox", "foobar")
    with k8s_instance() as instance:
        container_context = K8sContainerContext(
            image_pull_policy="IfNotPresent",
            env_secrets=["dagster-cloud-agent-token"],
            env_config_maps=["user-config-map"],
            service_account_name="dagster",
            image_pull_secrets=[{"name": "test-image-pull-secret"}],
            volume_mounts=[{"name": "foo", "mount_path": "biz/buz", "sub_path": "file.txt"}],
            volumes=[
                {"name": "foo", "config_map": {"name": "settings-cm"}},
            ],
            labels={"foo_label": "bar"},
            resources={
                "requests": {"cpu": "250m", "memory": "64Mi"},
                "limits": {"cpu": "500m", "memory": "2560Mi"},
            },
            scheduler_name="the_scheduler_name",
            server_k8s_config=UserDefinedDagsterK8sConfig.from_dict(
                {
                    "container_config": {
                        "name": "fooba",
                        "command": ["echo", "SERVER_OVERRIDE"],
                        "env": [{"name": "extra_key", "value": "extra_value"}],
                        "env_from": [{"config_map_ref": {"name": "that_config_map"}}],
                        "volume_mounts": [
                            {"name": "other_foo", "mount_path": "baz/boz", "sub_path": "foo.txt"}
                        ],
                        "resources": {
                            "requests": {"cpu": "500m", "memory": "128Mi"},
                            "limits": {"cpu": "250m", "memory": "5120Mi"},
                        },
                        "security_context": {"capabilities": {"add": ["SYS_PTRACE"]}},
                    },
                    "pod_template_spec_metadata": {
                        "namespace": "my_override_namespace",
                        "labels": {"my_other_label": "baz"},
                    },
                    "pod_spec_config": {
                        "dns_policy": "server_override_value",
                        "containers": [{"image": "sidecar_image", "name": "the_sidecar"}],
                        "image_pull_secrets": [{"name": "other-image-pull-secret"}],
                        "service_account_name": "other-service-account-name",
                        "volumes": [{"name": "other_foo", "config_map": {"name": "other-cm"}}],
                        "scheduler_name": "my_custom_scheduler_name",
                    },
                    "service_metadata": {"annotations": {"foo_service": "bar"}},
                    "service_spec_config": {"cluster_ip": "None"},
                    "deployment_metadata": {"annotations": {"foo_deployment": "baz"}},
                }
            ),
        )
        obj = construct_code_location_deployment(
            instance,
            deployment_name="sandbox",
            location_name="foobar_",
            k8s_deployment_name=resource_name,
            metadata=CodeLocationDeployData("bizbuz", package_name="blim"),
            container_context=container_context,
            args=["ls"],
            server_timestamp=time.time(),
        ).to_dict()

        assert obj

        deployment_metadata = obj["metadata"]
        assert deployment_metadata["annotations"] == {
            "foo_deployment": "baz",
        }

        pod_spec = obj["spec"]["template"]["spec"]
        assert pod_spec["dns_policy"] == "server_override_value"
        assert pod_spec["image_pull_secrets"][0]["name"] == "test-image-pull-secret"
        assert pod_spec["image_pull_secrets"][1]["name"] == "other-image-pull-secret"
        assert pod_spec["service_account_name"] == "other-service-account-name"
        assert pod_spec["scheduler_name"] == "my_custom_scheduler_name"

        assert pod_spec["volumes"][1]["name"] == "other_foo"

        containers = pod_spec["containers"]
        assert len(containers) == 2

        assert containers[1]["name"] == "the_sidecar"
        assert containers[1]["image"] == "sidecar_image"

        container = containers[0]

        assert container["name"] == "fooba"
        assert container["command"] == ["echo", "SERVER_OVERRIDE"]

        envs = {env["name"]: env["value"] for env in container["env"]}
        assert envs["extra_key"] == "extra_value"

        assert container["env_from"][2]["config_map_ref"]["name"] == "that_config_map"
        assert {
            "name": "other_foo",
            "mount_path": "baz/boz",
            "sub_path": "foo.txt",
        }.items() <= container["volume_mounts"][1].items()
        container["resources"].pop("claims", None)
        assert container["resources"] == {
            "requests": {"cpu": "500m", "memory": "128Mi"},
            "limits": {"cpu": "250m", "memory": "5120Mi"},
        }

        assert container["security_context"]["capabilities"]["add"] == ["SYS_PTRACE"]

        pod_metadata = obj["spec"]["template"]["metadata"]
        assert pod_metadata["namespace"] == "my_override_namespace"
        assert pod_metadata["labels"]["my_other_label"] == "baz"


def test_launch_k8s_server(kubeconfig_file):
    mock_k8s_apps_api_client = mock.MagicMock()
    mock_k8s_core_api_client = mock.MagicMock()
    with k8s_instance() as instance:
        user_code_launcher = K8sUserCodeLauncher(
            dagster_home="/opt/dagster/dagster_home",
            instance_config_map="dagster-instance",
            service_account_name="MY_SERVICE_ACCOUNT_NAME",
            namespace="default",
            kubeconfig_file=kubeconfig_file,
            k8s_apps_api_client=mock_k8s_apps_api_client,
            k8s_core_api_client=mock_k8s_core_api_client,
            server_k8s_config={
                "container_config": {"command": ["echo", "SERVER"], "tty": True},
                "pod_template_spec_metadata": {
                    "namespace": "my_server_namespace",
                    "labels": {"foo": "bar"},
                },
                "pod_spec_config": {"dns_policy": "server_value"},
                "deployment_metadata": {"annotations": {"foo_deployment": "bar_value"}},
                "service_metadata": {"annotations": {"foo_service": "bar_value"}},
                "service_spec_config": {"cluster_ip": "None"},
            },
            run_k8s_config={
                "container_config": {"command": ["echo", "RUN"]},
                "pod_template_spec_metadata": {"namespace": "my_run_namespace"},
                "pod_spec_config": {"dns_policy": "run_value"},
            },
            only_allow_user_defined_k8s_config_fields={
                "container_config": {
                    "command": True,
                },
                "pod_template_spec_metadata": {"labels": True},
                "pod_spec_config": {
                    "dns_policy": True,
                },
                "deployment_metadata": {"annotations": True},
                "service_metadata": {"annotations": True},
                "service_spec_config": {"cluster_ip": True},
                "namespace": True,
            },
            only_allow_user_defined_env_vars=["FOO", "BAR"],
        )
        user_code_launcher.register_instance(instance)  # pyright: ignore[reportArgumentType]

        user_code_launcher._start_new_server_spinup(  # noqa: SLF001
            deployment_name="acme",
            location_name="sandbox",
            desired_entry=UserCodeLauncherEntry(
                CodeLocationDeployData("bizbuz", package_name="blim"), time.time()
            ),
        )

        mock_method_calls = mock_k8s_apps_api_client.method_calls
        assert len(mock_method_calls) > 0
        method_name, _args, kwargs = mock_method_calls[0]
        assert method_name == "create_namespaced_deployment"

        body = kwargs["body"].to_dict()

        pod_spec = body["spec"]["template"]["spec"]
        pod_metadata = body["spec"]["template"]["metadata"]

        container = pod_spec["containers"][0]

        assert container["image"] == "bizbuz"
        assert container["command"] == ["echo", "SERVER"]

        assert pod_spec["dns_policy"] == "server_value"
        assert pod_metadata["namespace"] == "my_server_namespace"

        # Verify service creation includes service_spec_config
        service_method_calls = mock_k8s_core_api_client.method_calls
        service_method_name, service_args, _service_kwargs = service_method_calls[0]
        assert service_method_name == "create_namespaced_service"
        service_body = service_args[1].to_dict()
        assert service_body["spec"]["cluster_ip"] == "None"

        assert (
            user_code_launcher.run_launcher().run_k8s_config == user_code_launcher._run_k8s_config  # noqa: SLF001
        )

        # code server allowlist fields not included on run launcher
        assert user_code_launcher.run_launcher().only_allow_user_defined_k8s_config_fields == {
            "container_config": {"command": True, "env": True},
            "namespace": True,
            "pod_spec_config": {"dns_policy": True},
            "pod_template_spec_metadata": {"labels": True},
        }

        # Disallowed raw k8s fields are rejected
        with pytest.raises(
            Exception,
            match=r"Attempted to create a pod with fields that violated the allowed list: pod_template_spec_metadata.annotations",
        ):
            user_code_launcher._start_new_server_spinup(  # noqa: SLF001
                deployment_name="acme",
                location_name="sandbox",
                desired_entry=UserCodeLauncherEntry(
                    CodeLocationDeployData(
                        "bizbuz",
                        package_name="blim",
                        container_context={
                            "k8s": {
                                "server_k8s_config": {
                                    "pod_template_spec_metadata": {
                                        "annotations": {"foo_annot": "bar_annot"}
                                    }
                                }
                            }
                        },
                    ),
                    time.time(),
                ),
            )

        # Allowlisted raw k8s fields are allowed

        user_code_launcher._start_new_server_spinup(  # noqa: SLF001
            deployment_name="acme",
            location_name="sandbox",
            desired_entry=UserCodeLauncherEntry(
                CodeLocationDeployData(
                    "bizbuz",
                    package_name="blim",
                    container_context={
                        "env_vars": ["DAGSTER_CLOUD_LOCATION_NAME=sandbox"],
                        "k8s": {
                            "namespace": "my_override_namespace",
                            "server_k8s_config": {
                                "container_config": {
                                    "command": ["echo", "SERVER_OVERRIDE"],
                                    "env": [
                                        {"name": "FOO", "value": "FOO_VAL"},
                                        {"name": "BAZ", "value": "BAZ_VAL"},
                                        {"name": "BAR", "value": "BAR_VAL"},
                                    ],
                                },
                                "pod_template_spec_metadata": {
                                    "labels": {"baz": "quux"},
                                },
                                "pod_spec_config": {"dns_policy": "server_override_value"},
                            },
                        },
                    },
                ),
                time.time(),
            ),
        )

        # server_k8s_config values on the code location override the ones on the instance config

        mock_method_calls = mock_k8s_apps_api_client.method_calls
        method_name, _args, kwargs = mock_method_calls[1]
        assert method_name == "create_namespaced_deployment"

        namespace = kwargs["namespace"]
        assert namespace == "my_override_namespace"

        body = kwargs["body"].to_dict()

        pod_spec = body["spec"]["template"]["spec"]
        pod_metadata = body["spec"]["template"]["metadata"]

        container = pod_spec["containers"][0]

        assert container["command"] == ["echo", "SERVER_OVERRIDE"]
        assert container[
            "tty"
        ]  # still keeps keys from the instance config if they weren't supplied
        env_names = {env["name"] for env in container["env"]}
        assert {"FOO", "BAR", "DAGSTER_CLOUD_LOCATION_NAME"} <= env_names

        # "BAZ" excluded since it wasn't in the only_allow_user_defined_env_vars list
        assert "BAZ" not in env_names

        assert pod_spec["dns_policy"] == "server_override_value"
        assert pod_metadata["labels"]["baz"] == "quux"
        assert pod_metadata["labels"]["foo"] == "bar"

        # Ensure that agent_id was properly set on deployment
        deployment_labels = body["metadata"]["labels"]
        assert deployment_labels["agent_id"] == instance.instance_uuid  # pyright: ignore[reportAttributeAccessIssue]

        # Verify service_spec_config from location-level container_context is applied
        # This exercises the persisted code-location container_context path
        # (K8sContainerContext.create_from_config -> job.py config_type_container_context)

        user_code_launcher._start_new_server_spinup(  # noqa: SLF001
            deployment_name="acme",
            location_name="sandbox",
            desired_entry=UserCodeLauncherEntry(
                CodeLocationDeployData(
                    "bizbuz",
                    package_name="blim",
                    container_context={
                        "k8s": {
                            "namespace": "my_override_namespace",
                            "server_k8s_config": {
                                "container_config": {
                                    "command": ["echo", "SERVER_OVERRIDE"],
                                },
                                "service_spec_config": {"cluster_ip": "None"},
                            },
                        }
                    },
                ),
                time.time(),
            ),
        )

        # Verify the service was created with service_spec_config from location context
        service_calls = [
            call
            for call in mock_k8s_core_api_client.method_calls
            if call[0] == "create_namespaced_service"
        ]
        latest_service = service_calls[-1]
        service_body = latest_service[1][1].to_dict()
        assert service_body["spec"]["cluster_ip"] == "None"

        # Create a server with merge_behavior SHALLOW - labels are replaced rather than added

        user_code_launcher._start_new_server_spinup(  # noqa: SLF001
            deployment_name="acme",
            location_name="sandbox",
            desired_entry=UserCodeLauncherEntry(
                CodeLocationDeployData(
                    "bizbuz",
                    package_name="blim",
                    container_context={
                        "k8s": {
                            "namespace": "my_override_namespace",
                            "server_k8s_config": {
                                "container_config": {
                                    "command": ["echo", "SERVER_OVERRIDE"],
                                },
                                "pod_template_spec_metadata": {
                                    "labels": {"baz": "quux"},
                                },
                                "pod_spec_config": {"dns_policy": "server_override_value"},
                                "merge_behavior": "SHALLOW",
                            },
                        }
                    },
                ),
                time.time(),
            ),
        )
        mock_method_calls = mock_k8s_apps_api_client.method_calls
        method_name, _args, kwargs = mock_method_calls[-1]
        assert method_name == "create_namespaced_deployment"

        body = kwargs["body"].to_dict()

        pod_spec = body["spec"]["template"]["spec"]
        pod_metadata = body["spec"]["template"]["metadata"]
        assert pod_metadata["labels"]["baz"] == "quux"
        assert "foo" not in pod_metadata["labels"]


@pytest.fixture
def core_api_client(mocker):
    return mocker.Mock()


@pytest.fixture
def apps_api_client(mocker):
    api_client = mocker.Mock()
    replicaset_mock = Mock()
    replicaset_mock.metadata.name = "test-replicaset"
    api_client.list_namespaced_replica_set.return_value.items = [replicaset_mock]
    return api_client


@pytest.fixture
def mock_dagster_k8s_client():
    with mock.patch(
        "dagster_cloud.workspace.kubernetes.utils.DagsterKubernetesClient.production_client"
    ) as mock_dagster_k8s_client:
        api_client_mock = Mock()
        api_client_mock.get_pod_debug_info.return_value = "Pod debug info"
        mock_dagster_k8s_client.return_value = api_client_mock
        yield mock_dagster_k8s_client


def test_get_deployment_failure_debug_info_no_replicaset_warnings(
    core_api_client, apps_api_client, mock_dagster_k8s_client
):
    namespace = "default"
    k8s_deployment_name = "test-deployment"

    core_api_client.list_namespaced_event.return_value.items = []

    pod_mock = Mock()
    pod_mock.metadata.name = "test-pod"
    pod_list = [pod_mock]

    result = get_deployment_failure_debug_info(
        k8s_deployment_name,
        namespace,
        core_api_client,
        pod_list,
        logging.getLogger("test"),
        apps_api_client,
    )

    assert (
        result
        == """Pod debug info

No warning events for replicaset test-replicaset.

For more information about the failure, run `kubectl describe pod test-pod` or `kubectl describe deployment test-deployment` in your cluster."""
    )


def test_get_deployment_failure_debug_info_with_replicaset_warnings(
    core_api_client, apps_api_client, mock_dagster_k8s_client
):
    namespace = "default"
    k8s_deployment_name = "test-deployment"

    event_mock = Mock()
    event_mock.reason = "FailedCreate"
    event_mock.message = "Error creating: pod already exists"
    event_mock.count = 10
    core_api_client.list_namespaced_event.return_value.items = [event_mock]

    pod_mock = Mock()
    pod_mock.metadata.name = "test-pod"
    pod_list = [pod_mock]

    result = get_deployment_failure_debug_info(
        k8s_deployment_name,
        namespace,
        core_api_client,
        pod_list,
        logging.getLogger("test"),
        apps_api_client,
    )

    assert (
        result
        == """Pod debug info

Warning events for replicaset test-replicaset:
FailedCreate: Error creating: pod already exists (x10)

For more information about the failure, run `kubectl describe pod test-pod` or `kubectl describe deployment test-deployment` in your cluster."""
    )


def test_get_deployment_failure_debug_info_with_exceptions(
    core_api_client, apps_api_client, caplog
):
    namespace = "default"
    k8s_deployment_name = "test-deployment"

    apps_api_client.list_namespaced_replica_set.side_effect = Exception("Replicaset error")

    core_api_client.list_namespaced_event.return_value.items = []

    pod_list = []

    with caplog.at_level(logging.ERROR, logger="test"):
        result = get_deployment_failure_debug_info(
            k8s_deployment_name,
            namespace,
            core_api_client,
            pod_list,
            logging.getLogger("test"),
            apps_api_client,
        )

        assert (
            result
            == "For more information about the failure, run `kubectl describe deployment test-deployment` in your cluster."
        )

        records = [r for r in caplog.records if r.name == "test"]
        assert len(records) == 1
        assert "Failure fetching replicaset debug info" in str(records[0])


def test_construct_code_location_service_custom_port(monkeypatch):
    """Test that construct_code_location_service uses DAGSTER_CLOUD_CODE_SERVER_PORT env var."""
    monkeypatch.setenv("DAGSTER_CLOUD_CODE_SERVER_PORT", "5000")

    resource_name = unique_k8s_resource_name("sandbox", "biz.buz")
    container_context = K8sContainerContext()

    with k8s_instance() as instance:
        server_timestamp = time.time()
        obj = construct_code_location_service(
            "sandbox",
            "biz.buz",
            resource_name,
            container_context,
            instance,
            server_timestamp,
        ).to_dict()

        assert obj["spec"]["ports"][0]["port"] == 5000


def test_construct_code_location_deployment_custom_port(monkeypatch):
    """Test that construct_code_location_deployment uses DAGSTER_CLOUD_CODE_SERVER_PORT env var."""
    monkeypatch.setenv("DAGSTER_CLOUD_CODE_SERVER_PORT", "5000")

    resource_name = unique_k8s_resource_name("sandbox", "foobar")
    container_context = K8sContainerContext()

    with k8s_instance() as instance:
        obj = construct_code_location_deployment(
            instance,
            deployment_name="sandbox",
            location_name="foobar_",
            k8s_deployment_name=resource_name,
            metadata=CodeLocationDeployData("bizbuz", package_name="blim"),
            container_context=container_context,
            args=["ls"],
            server_timestamp=time.time(),
        ).to_dict()

        container_env = obj["spec"]["template"]["spec"]["containers"][0]["env"]
        port_env = next(e for e in container_env if e["name"] == "DAGSTER_CLI_API_GRPC_PORT")
        assert port_env["value"] == "5000"
